#!/usr/bin/env python
import os
import warnings
import shelve
import pickle
import datetime
import shlex

import numpy as np
from scipy import interpolate
from scipy import stats
import matplotlib.pyplot as plt

from vg.meteo import meteox2y, avrwind
from vg import times
from vg.time_series_analysis import time_series as ts
from vg.time_series_analysis import models
from vg.time_series_analysis import distributions
import vg.time_series_analysis.seasonal_distributions as sd
import vg.time_series_analysis.seasonal_kde as skde
from vg import helpers as my
try:
    import config as conf
except ImportError:
    from vg import config_template as conf
    conf_filepath = conf.__file__
    if conf_filepath.endswith(".pyc"):
        conf_filepath = conf_filepath[:-1]
    warnings.warn('Could not import "config.py".\n' +
           'Edit "%s" and rename it to "config.py"' % conf_filepath)


def var_names_greek(var_names):
    return [conf.ygreek[var_name] for var_name in var_names]


def _parse_time(var_dict):
    """Get the time information in var_dict. This is adjusted every time
    someone throws new data at me."""
    if "Julian" in var_dict:
        datetimes = times.cwr2datetime(var_dict["Julian"])
        del var_dict["Julian"]
    elif "jd" in var_dict:
        datetimes = times.cwr2datetime(var_dict["jd"])
        del var_dict["jd"]
    if "chron" in var_dict:
        datetimes = times.str2datetime(var_dict["chron"],
                                       "(%m/%d/%y %H:%M:%S)")
        del var_dict["chron"]
    elif "date" in var_dict:
        datetimes = times.str2datetime(var_dict["date"])
        del var_dict["date"]
    elif "time" in var_dict:
        datetimes = times.iso2datetime(var_dict["time"])
        del var_dict["time"]
    if "year" in var_dict and "doy" in var_dict:
        datetimes = times.str2datetime(var_dict["year"], "%Y")
        datetimes += np.array([datetime.timedelta(int(doy))
                               for doy in var_dict["doy"]])
        del var_dict["year"], var_dict["doy"]
    if "Date" in var_dict and "Time" in var_dict:
        # as seen in Excel
        datetimes = times.xls2datetime([int(day) for day in var_dict["Date"]])
        datetimes += \
            np.array([datetime.timedelta(hours=int(hour.split(":")[0]))
                      for hour in var_dict["Time"]])
        del var_dict["Date"], var_dict["Time"]
    if "ta" in var_dict and "mo" in var_dict and "jahr" in var_dict:
        datetimes = \
            np.array([datetime.datetime(int(year), int(month), int(day))
                      for year, month, day in
                      zip(var_dict["jahr"], var_dict["mo"], var_dict["ta"])])
        del var_dict["ta"]
        del var_dict["mo"]
        del var_dict["jahr"]
    if "index" in var_dict:
        del var_dict["index"]
    return var_dict, datetimes


@my.pickle_cache(os.path.join(conf.cache_dir, "met_%s.pkl"), warn=False)
def read_met(filepath=None, minimum_water_temp=.1, delimiter="\t",
             verbose=True, **kwds):
    if filepath is None:
        filepath = conf.met_file
    var_dict = my.csv2dict(filepath, delimiter=delimiter, **kwds)
    # get rid of quotes around filenames
    var_dict = {shlex.split(var_name)[0]: values
                for var_name, values in var_dict.items()}

    var_dict, datetimes = _parse_time(var_dict)

    met = {key: np.array([val if val != "" else np.nan for val in values],
                         dtype=float)
           for key, values in var_dict.items()}

    met = {key: times.regularize(val, datetimes)[0]
           for key, val in met.items()}
    # get also the regularized datetimes
    datetimes = times.regularize(met[met.keys()[0]], datetimes)[1]

    if verbose:
        n_nan = np.sum([np.isnan(val) for val in met.values()])
        print "\tFound %d missing or nan values" % n_nan
        for var_name, val in sorted(met.items()):
            nan_mask = np.isnan(val)
            if np.sum(nan_mask) > 0:
                print "\t", var_name
                for ii in np.where(nan_mask)[0]:
                    print "\t\t", datetimes[ii]

    if "Cloud_Cover" not in var_dict:
        try:
            var_dict["Cloud_Cover"] = \
                meteox2y.lw2clouds(np.array(var_dict["ILWR"], dtype=float),
                                   np.array(var_dict["theta"], dtype=float),
                                   e=np.array(var_dict["e"], dtype=float))
        except KeyError:
            pass

    if "wdir" in met and "U" in met:
        u, v = avrwind.angle2component(met["wdir"], met["U"])
        met.update({"u": u, "v": v})

    if ("rh" not in met) and ("e" in met) and ("theta" in met):
        met["rh"] = meteox2y.vap_p2rel(met["e"], met["theta"])
        too_dry_ii = met["rh"] < 0
        n_too_dry = np.sum(too_dry_ii)
        if n_too_dry > 0:
            if verbose:
                warnings.warn(("Input of e and theta caused %d rel. " +
                               "humidities to be < 0. Capping them at 0.")
                              % n_too_dry)
            met["rh"][too_dry_ii] = np.nan
        too_moist_ii = met["rh"] > 1
        n_too_moist = np.sum(too_moist_ii)
        if n_too_moist > 0:
            if verbose:
                warnings.warn(("Input of e and theta caused %d rel. " +
                               "humidities to be > 1. Capping them at 1.")
                              % n_too_moist)
            met["rh"][too_moist_ii] = np.nan
    elif "rh" in met and met["rh"].max() > 50:
        met["rh"] /= 100
    for var_name in met.keys():
        if var_name.startswith("wtemp"):
            too_cold_ii = met[var_name] < 0
            n_too_cold = np.sum(too_cold_ii)
            if n_too_cold > 0:
                if verbose:
                    warnings.warn(("%d temperatures < 0 in %s. Capping them " +
                                   "at %.3f")
                                  % (n_too_cold, var_name, minimum_water_temp))
                met[var_name][too_cold_ii] = minimum_water_temp

    return datetimes, met


def plot_monthly_autocorr(times, met, var_names=None, k_range=10,
                          n_sumup=24, **kwds):
    if var_names is None:
        var_names = conf.dists.keys()
    var = met_as_array(met, var_names=var_names)
    if n_sumup > 1:
        var, times = my.sumup(var, 24, times, middle_time=False)
    for month in range(1, 13):
        values = var[:, times.time_part(times, "%m") == month]
        fig = ts.plot_auto_corr(values, k_range=k_range, var_names=var_names,
                                n_difference=0, **kwds)
        fig.canvas.set_window_title("%s (%d)" %
                                    (str(month), fig.canvas.manager.num))


def plot_monthly_corrcoeff(times, met, var_names=None, n_sumup=24, **kwds):
    if var_names is None:
        var_names = conf.dists.keys()
    var = met_as_array(met, var_names=var_names)
    if n_sumup > 1:
        var, times = my.sumup(var, 24, times, middle_time=False)
    for month in range(1, 13):
        values = var[:, times.time_part(times, "%m") == month]
        fig = ts.corr_img(values, title=str(month), var_names=var_names,
                          **kwds)
        fig.canvas.set_window_title("%s (%d)" %
                                    (str(month), fig.canvas.manager.num))


def plot_time_hists(data, times, title_template, var_names, ylabels=None,
                    all_var_names=None, window_title=True, figsize=None, *args,
                    **kwds):
    if ylabels is None:
        ylabels = len(var_names) * [None]
    if type(var_names) is str:
        var_names = (var_names,)
    figs = []
    for var_name in var_names:
        var_ii = all_var_names.index(var_name)
        values = data[var_ii]
        ylabel = ylabels[var_ii]
        fig = my.time_hist(values[np.isfinite(values)],
                           times[np.isfinite(values)], ylabel=ylabel,
                           figsize=figsize, trend=False, *args, **kwds)
        title_str = (title_template %
                     ("%s %s" % (conf.long_var_names[var_name],
                                 conf.ygreek[var_name])
                      if var_name in conf.long_var_names else var_name))
        if "[" in title_str:
            title_str = title_str.split("[")[0]
        plt.suptitle(title_str)  # + " mean:%.2f" % values.mean())
        if window_title:
            fig.canvas.set_window_title("%s (%d)" %
                                        (title_str, fig.canvas.manager.num))
        figs.append(fig)
    return figs


def plot_copulas(data, var_names, title=None, lag=0, **fig_kwds):
    if lag == 0:
        fig = my.cplom(data, var_names, title=title, **fig_kwds)
    else:
        fig = my.ccplom(data, lag, var_names, title=title, **fig_kwds)
    if title:
        fig.canvas.set_window_title("%s (%d)" %
                                    (title, fig.canvas.manager.num))


def psd_plots(arrays, var_names, categories=("raw", "transformed",
                                             "simulated", "backtransformed")
              ):
    for var_ii in range(len(var_names)):
        fig = plt.figure()
        for category_ii in range(len(categories)):
            # TODO: sampling frequency is fixed here at 1 per day
            plt.psd(arrays[category_ii][var_ii], Fs=1., figure=fig,
                    scale_by_freq=False)
        var_name = var_names[var_ii]
        plt.title(var_name)
        plt.legend(categories)
        xtick_locs = plt.xticks()[0]
        xtick_labels = ["%.2f" % (1. / loc) for loc in xtick_locs]
        plt.xticks(xtick_locs, xtick_labels)
        plt.xlabel("wavelength in days")
        fig.canvas.set_window_title("%s power spectrum density (%d)" %
                                    (var_name, fig.canvas.manager.num))


def daily_wind(met, sum_interval=24):
    """Aggregates daily wind direction and speed from hourly values."""
    u_daily = my.sumup(met["u"], sum_interval)
    v_daily = my.sumup(met["v"], sum_interval)
    return avrwind.component2angle(u_daily, v_daily)


def phi_main(u, v):
    w, ve = np.linalg.eig(np.cov(u, v))
    v_main = ve[:, np.argmax(w)]
    return np.arctan2(v_main[1], v_main[0])


def turn_uv(u, v, phi):
    uv = np.array([u, v]).T
    turn_matrix = np.matrix([[np.cos(phi), -np.sin(phi)],
                             [np.sin(phi), np.cos(phi)]])
    uv = uv * turn_matrix
    u, v = np.asarray(uv).T
    return np.squeeze(u), np.squeeze(v)


def detrend(values):
    """Detrend by subtracting the linear regression function."""
    dummy_time = np.arange(len(values))
    b, a = stats.linregress(dummy_time, values)[:2]
    return values + (values.mean() - (a + b * dummy_time))


def dump_data(times, data, var_names, p, q, extra="", random_state=None):
    """Dumps the given datetimes and data in an ascii file and random state in
    a pickle file."""
    outfilename = "VARMA_p%d_q%d_sim%s" % (p, q, extra)
    outfilepath = os.path.join(conf.out_dir, outfilename)
    txt_file = open(outfilepath + ".dat", "w")
    txt_file.write("time\t" + "\t".join(var_names) + "\n")
    for time, values in zip(times, data.T):
        txt_file.write("%s\t%s\n" % (time.isoformat(),
                        "\t".join("%.3f" % val for val in values)))
    txt_file.close()

    if random_state is not None:
        with open(outfilepath + ".random_state", "wb") as pi_file:
            pickle.dump(random_state, pi_file)


def met_as_array(met, p=0, T=None, var_names=None):
    """Returns the met - dictionary as an array."""
    var_names = sorted(met.keys()) if var_names is None else var_names
    K = len(var_names)
    if T is None:
        T = len(met[var_names[0]]) - p
    # this lets some outer code access the following variables without
    # returning them explicitly
    met_as_array.var_names, met_as_array.T = var_names, T
    # the "T + p" is there to include the presample timesteps.
    data = np.nan * np.empty((K, T + p))
    for var_i, key in enumerate(var_names):
        data[var_i] = met[key]
    return data


def seasonal_back(dist_sol, norm_data, var_names, doys=None):
    data = np.empty_like(norm_data)
    for var_ii, (var_name, var) in enumerate(zip(var_names, norm_data)):
        distribution, solution = dist_sol[var_name]
        # TODO: distribution is never distributions.norm. however the transform
        # yields wrong results!
        if distribution is distributions.norm:
#        if (hasattr(distribution, "dist") and
#            type(distribution.dist) is distributions.Normal):
            # in this case, we do not need a qenuine qq-transform, the normal
            # inverse Z transform suffices, AND is immune to producing nans
            # for "extreme" values in the standard-normal world!
            if doys is not None:
                _T = (2 * np.pi / 365 * doys)[np.newaxis, :]
            sigmas = distribution.trig2pars(solution, _T)[1]
            data[var_ii] *= sigmas
        else:
            quantiles = distributions.norm.cdf(var, 0, 1)
            data[var_ii] = distribution.ppf(solution, quantiles, doys=doys)
    return data


def sim_byvariable(values, times, p=2, aggregation="%w"):
    aggr_values, aggr_times = \
        ts.time_part_average(values, times, aggregation, expand=False,
                          return_times=True)
    dist = sd.SeasonalDist(distributions.norm, aggr_values, aggr_times)
    solution = dist.fit(T0=50)
    trans = distributions.norm.ppf(dist.cdf(solution), 0, 1)
    B, sigma_u = models.VAR_LS(trans, p=p)
    T = len(aggr_values)
    sim = models.VAR_LS_sim(B, sigma_u, T - p)
    retrans = dist.ppf(solution, distributions.norm.cdf(sim, 0, 1))
    retrans = np.repeat(retrans, times.time_part_average.repeats)
    return retrans


def interannual_variability(T, mean_arrival=30, disturbance_std=.1):
    """Generate a time series used as a disturbance within the std-norm world.
    Disturbances are simulated as gaussian changes occuring after t timesteps,
    where t is drawn from an exponential distribution.
    """
    arrival = \
        lambda: distributions.expon.ppf(np.random.rand(),
                                        lambd=1. / mean_arrival,
                                        # loc=1 weil wir taegliche Werte
                                        # berechnen und daher 1 Tag die
                                        # kuerzest moegliche Dauer ist
                                        x0=1)
    disturbance = \
        lambda: abs(distributions.norm.ppf(np.random.rand(),
                                           sigma=disturbance_std))
    m_t = np.zeros(T)
    breaks_ii = [t] = [int(round(arrival()))]
    while t < T:
        breaks_ii += [breaks_ii[-1] + int(round(arrival()))]
        t = breaks_ii[-1]
    sign = 1
    for break_, next_break in zip(breaks_ii[:-1], breaks_ii[1:]):
        period_length = min(int(round(next_break - break_)),
                            len(m_t) - break_)
        window = np.bartlett(period_length + 1)[:-1]
        if window.mean() <= 0:
            window = np.ones(period_length) / period_length
        else:
            window /= window.mean()
        m_t[break_:next_break] = sign * disturbance() * window
        sign *= -1
    return m_t


def sim_times(T, resolution, start_date):
    """Generates two arrays:
        1. datetimes starting at start_date with T values and resolution
           hours in between.
        2. the days-of-year of 1.

    Parameters
    ----------
    T : int
        number of timesteps
    resolution : int
        length of timestep in hours
    start_date : datetime object
        Representation of first timestep.

    Returns
    -------
    times_out : (T,) ndarray, dtype=object (datetime)
    sim_doys : (T,) ndarray, dtype=int
        Day of years of `times_out`

    Examples
    --------
    >>> import datetime
    >>> sim_times(4, 48, datetime.datetime(year=2000, month=1, day=1))
    (array([2000-01-01 00:00:00, 2000-01-03 00:00:00, 2000-01-05 00:00:00,
           2000-01-07 00:00:00], dtype=object), array([1, 3, 5, 7]))
    """
    interval_secs = 60. ** 2 * resolution
    times_out = np.cumsum([0] + (T - 1) * [interval_secs])
    times_out += times.datetime2unix(start_date)
    times_out = times.unix2datetime(times_out)
    return times_out, times.datetime2doy(times_out)


def sw_daily(date, daily_sw_data, del_t=3600):
    """ add standard daily cycle to daily mean sw data
    Parameters
    ----------
    date: array
        date in unix timestamps
    daily_sw_data: array
    del_t: int
        timestep of output data in seconds, default = 3600s

    Returns
    -------
    date_: array
        output date in unix timestamps
    data_: array
        sw data in del_t-steps"""
    nn = 86400. / del_t
    date_ = np.arange(date[0], date[-1] + 86400, del_t)
    swmax = meteox2y.pot_s_rad(times.unix2str(date_, "%Y-%m-%dT%H:%M"),
                               lat=47.633, longt=9.367)
    swmax_daily = np.average(swmax.reshape(-1, nn), axis=1).repeat(nn)
    fkt = swmax / swmax_daily
    data_ = daily_sw_data.repeat(nn) * fkt
    return date_, data_


class VG(object):
    """A Vector-Autoregressive weather generator.

    >>> my_vg = VG(("theta", "Qsw", "ILWR", "rh", "u", "v"))
    >>> my_vg.fit()
    >>> times_out, sim_data = my_vg.simulate()
    """
    def __init__(self, var_names, met_file=None, sum_interval=24, plot=False,
                 separator="\t", refit=None, detrend_vars=None, verbose=False,
                 data_dir=None, cache_dir=None, **met_kwds):
        """A vector autoregressive moving average (VARMA) weather generator.
        Initializing a VG-object does the following:
            - read input data from a met-file
            - aggregate it according to sum_interval
            - convert the data using seasonal distributions

        To fit the VARMA process call fit(). Afterwards
        simulated time series can be obtained with simulate().

        Parameters
        ----------
        var_names : sequence containing strings
        met_file : string or None, optional
            If None, value will be read from config.py
        sum_inverval : int or sequence of ints, optional
            Number of timesteps to aggregate. If a sequence is given, its
            elements are used variable - specific. In this case the sequence
            has to have the same length as var_names.
            Default 24 corresponds to the case of hourly input data and daily
            simulation
        plot : boolean, optional
            Plot all kinds of information. Warning: a lot of windows will
            appear.
        separator : str or None, optional
            String that separates columns in the met_file. It is fed to the
            split method of str, so None corresponds to any whitespace.
        refit : None, 'all' or sequence of strings from var_names, optional
            Instead of using cached solutions for the seasonal distribution
            fit, run the fit again and cache it.
        detrend_vars : sequence containing strings or None, optional
            Remove a linear trend from the variables.
        verbose : boolean, optional
            Be more verbose if True.
        data_dir : str, optional
            Overwrites the data_dir variable from the config-file.
        cache_dir : str, optional
            Overwrites the cache_dir variable from the config-file.
        """
        self.var_names = var_names
        self.verbose = verbose
        self.met_file = conf.met_file if met_file is None else met_file
        self.data_dir = conf.data_dir if data_dir is None else data_dir
        if cache_dir is not None:
            self.cache_dir = cache_dir
            self.seasonal_cache_file = os.path.join(cache_dir,
                                                    "seasonal_solutions.she")
        else:
            self.cache_dir = conf.cache_dir
            self.seasonal_cache_file = conf.seasonal_cache_file
        # will be used in _load_and_prepare_data
        self.met = self.times_orig = None
        # are we given a sequence? then store sum_interval in a way, that
        # data can be divided by it
        try:
            self.sum_interval = np.array(sum_interval)[:, np.newaxis]
        except IndexError:
            # let's make this the same shape, so life gets easier later on
            self.sum_interval = \
                np.array(self.K * [sum_interval])[:, np.newaxis]
        self.plot = plot

        # attributes that will be set with meaningfull values in fit()
        # AM is also used to store the parameters of a pure VAR-process
        self.AM = self.sigma_u = self.residuals = self.means = None
        self.p = self.q = None

        # these attributes will be populated in simulate()
        self.T = self.mean_arrival = self.disturbance_std = None
        self.theta_incr = self.theta_grad = self.fixed_variables = None
        self.primary_var = self.climate_signal = self.start_date = None
        self.m_t = self.sim = self.sim_sea = self.sim_times = None

        print "Loading input data."
        self.times, self.data_raw = self._load_and_prepare_data(separator,
                                                                **met_kwds)

        if plot:
            ts.corr_img(self.data_raw / self.sum_interval, 0,
                     "Correlation matrix of raw data", var_names)

        if detrend_vars:
            for var_name in detrend_vars:
                var_ii = var_names.index(var_name)
                self.data_raw[var_ii] = detrend(self.data_raw[var_ii])

        # fit seasonal distributions
        print "Fitting seasonal distributions."
        self.data_trans, self.dist_sol = self._fit_seasonal(refit)

        if plot:
            ts.corr_img(self.data_trans, 0,
                     "Correlation matrix of transformed data", var_names)
            plot_copulas(self.data_trans, var_names,
                         title="Copula plot matrix of transformed data")

    @property
    def ylabels(self):
        return [conf.units[var] if var in conf.units else "[?]"
                for var in self.var_names]

    @property
    def sum_interval_dict(self):
        """Maps var_name to sum_interval."""
        return dict((var_name,
                     self.sum_interval[self.var_names.index(var_name)])
                    for var_name in self.var_names)

    @property
    def primary_var_ii(self):
        """The row index of the primary variable."""
        try:
            return self.var_names.index(self.primary_var)
        except ValueError:
            warnings.warn("No %s in input." % self.primary_var)
            return None

    @property
    def output_resolution(self):
        """In hours."""
        delta = self.times[1] - self.times[0]
        # there is timedelta.total_seconds in python 2.7, but lets be kind to
        # the conservative people -- oh, thank you ;-)
        return (24 * delta.days +
                (delta.seconds + 1e-3 * delta.microseconds) / 60 ** 2)

    @property
    def K(self):
        """The number of variables."""
        return len(self.var_names)

    @property
    def T_data(self):
        """Length of the input data time series."""
        return len(self.times_orig)

    @property
    def T_summed(self):
        """Length of aggregated input data."""
        return len(self.times)

    @property
    def data_doys(self):
        """Days of year of aggregated input data."""
        return times.datetime2doy(self.times)

    @property
    def sim_doys(self):
        """Days of year of aggregated input data."""
        return times.datetime2doy(self.sim_times)

    def fit(self, p=None, q=0, p_max=10):
        """p is the order of autoregressive process.
        q the order of the moving average process
        If p is not given, it is selected using the Schwartz Information
        Criterion."""
        self.p, self.q = p, q
        if q is 0:
            if p is None:
                print "Perfoming VAR order selection: ",
                try:
                    self.p = p = \
                        models.VAR_order_selection(self.data_trans, p_max)
                except ValueError:
                    warnings.warn("Could not fit the VAR process. Trying to "
                                  "remove extreme values")
                    mi, ma = np.min(self.data_trans), np.max(self.data_trans)
                    if ma > -mi:
                        self.data_trans.ravel()[np.argmax(self.data_trans)] = \
                            np.nan
                    else:
                        self.data_trans.ravel()[np.argmin(self.data_trans)] = \
                            np.nan
                    self.data_trans = my.interp_nan(self.data_trans)
                    self.fit(p, q, p_max)
                print "p=%d seems parsimonious to me" % self.p
                if p == 0:
                    print "Wow, you should not do any autoregressive modeling."
            ## fit a VAR-model
            print "Fitting the VAR model"
            B, self.sigma_u = models.VAR_LS(self.data_trans, self.p)
            # we call it AM to have one place for VARMA parameters
            self.AM = B
            self.residuals = models.VAR_residuals(self.data_trans, B, self.p)

        else:
            if (p is None) and (q is None):
                print "Perfoming VARMA order selection: ",
                self.p, self.q = p, q = \
                    models.VARMA_order_selection(self.data_trans, p_max,
                                                 plot_table=self.plot)
                print "p=%d and q=%d seems parsimonious to me" % (p, q)
            else:
                self.p, self.q = p, q
            # fit a VARMA-model
            print "Fitting the VARMA model"
            # the VARMA implementation is for mean-adjusted data. the data
            # should have zero mean, asuming the fit for the qq-transformation
            # was perfect.
            # ... and nothing is perfect. so we subtract the mean here
            self.means = self.data_trans.mean(axis=1)[:, np.newaxis]
            self.data_trans -= self.means

            AM, sigma_u, residuals = \
                models.VARMA_LS_prelim(self.data_trans, p, q)

            self.data_trans += self.means
            self.AM, self.sigma_u, self.residuals = AM, sigma_u, residuals

        if self.plot:
            ts.matr_img(np.asarray(self.AM), "AM p=%d q=%d" % (p, q), None,
                     self.var_names)

    def simulate(self, T=None, mean_arrival=7, disturbance_std=0,
                 theta_incr=0, theta_grad=0, fixed_variables=None,
                 primary_var="theta", climate_signal=None, start_str=None,
                 random_state=None):
        """Simulate based on data from __init__ and the fit from calling
        self.fit().
        To choose a specific order of AR and MA call fit(p, q) before calling
        simulate. Otherwise fit will be called without parameters, so p and q
        will be guessed by applying an information criterion.

        Parameters
        ----------
        T : None or int, optional
            Number of timesteps to simulate. T=None produces as many simulation
            timesteps as there were in the data used for fitting the VARMA
            process.
        mean_arrival : float, optional
            average length of warm/cold periods [d]
        disturbance_std : float, optional
            standard deviation of deviation of warm/cold periods from mean
        theta_incr : float, optional
            increase in average of temperature [deg C]
        theta_grad : float, optional
            gradient increase in average of temperature [deg C / (T)]
        fixed_variables : dictionary mapping var_names to None or (T,) ndarrays
            Keeps the provided time - series fixed. Mappings to None indicate
            that the input data should be used as a fixed variable.
            Can be used to simulate hierarchically.
        primary_var : str, one of var_names from __init__, optional
            All disturbances (mean_arrival, distrubance_std, theta_incr,
            theta_grad and climate_signal) correspond to changes in this
            variable.
            !!! This must be a variable that was fitted with a normal
            distribution!!!
        climate_signal : (T,) ndarray
            A time series of the 'primary_var'. Differences between
            'climate_signal' and the seasonal mean are used to perturb the
            VAR(MA) process.
        start_str : str of format '%m-%d %H', optional
            String representation of the start date of the simulation.
        random_state : str, tuple as returned by numpy.random.get_state() or
                       None, optional
            Sets the random state before simulating.
            *str* is interpreted as a pickle file containing the random state.
            *tuple* is interpreted as the random state itself.

        Returns
        -------
        times_out : (T,) ndarray of datetime objects
        sim_sea : (K, T) float ndarray
        """

        # check if primary variable is normally distributed
        primary_var_dist = self.dist_sol[primary_var][0]
        if type(primary_var_dist.dist) is not distributions.Normal:
            warnings.warn("Primary variable %s has to be normally distributed!"
                          % primary_var +
                          "%s was used here.\nAborting." % primary_var_dist)
            return None, None

        if type(random_state) is str:
            with open(random_state, "rb") as pi_file:
                random_state = pickle.load(pi_file)
        if random_state is not None:
            np.random.set_state(random_state)
        # we record the random state to later dump it
        pre_sim_random_state = np.random.get_state()

        self.T, self.mean_arrival = T, mean_arrival
        self.disturbance_std, self.theta_incr = disturbance_std, theta_incr
        self.theta_grad, self.fixed_variables = theta_grad, fixed_variables
        self.primary_var, self.climate_signal = primary_var, climate_signal

        # we depend on order selection if self.fit was not called
        if self.p is None:
            self.fit()

        if T is None:
            if climate_signal is not None:
                self.T = len(climate_signal)
            else:
                self.T = self.T_summed
        else:
            self.T = T

        # produce sim_doys and times_out:
        if start_str is None:
            self.start_date = self.times_orig[0]
        else:
            self.start_date = times.str2datetime(start_str)
        self.sim_times, sim_doys = sim_times(self.T, self.output_resolution,
                                             self.start_date)
        # converts the given parameters so they are understood in the
        # std-normal world
        S, m, m_t, m_trend = self._scenario_parameters(sim_doys)
        # store to be able to play outside with it later
        self.m_t = m_t

        fixed_data = self._prepare_fixed_data()

        print "Simulating a time-series."
        if self.q in (0, None):
            ## simulate VAR-time-series
            sim = models.VAR_LS_sim(self.AM, self.sigma_u, self.T,
                                    m, ia=m_t, m_trend=m_trend,
                                    fixed_data=fixed_data)
        else:
            sim = models.VARMA_LS_sim(self.AM, self.p, self.q, self.sigma_u,
                               # process means that are used as starting values
                               # there is some confusion here...
                               self.means,
                               self.T, S, m, ia=m_t, m_trend=m_trend,
                               fixed_data=fixed_data)

        if self.plot:
            ts.matr_img(self.sigma_u, "Noise Covariance matrix",
                        self.var_names, self.var_names)
            ts.corr_img(sim, 0, "Correlation matrix of simulated data",
                     self.var_names)
            plot_copulas(sim, self.var_names,
                         title="Copula plot matrix of simulated data")

        # transform back
        sim_sea = seasonal_back(self.dist_sol, sim, self.var_names,
                                doys=sim_doys)
        sim_sea /= self.sum_interval

        # checking the change of moments
        try:
            mean_before = (self.data_raw[self.primary_var_ii] /
                           self.sum_interval[self.primary_var_ii]).mean()
            mean_after = my.nanavg(sim_sea[self.primary_var_ii])
#            print ("\t%d NaN-Werte bei %s" %
#                   (np.sum(~np.isfinite(sim_sea[self.primary_var_ii])),
#                   conf.long_var_names[self.primary_var]))
            print (("\t%s mean (data): %.4f, %s mean (sim): %.4f, "
                    "diff: %.4f")  # , mean(m): %.4f")
                   % (self.primary_var, mean_before, self.primary_var,
                      mean_after, mean_after - mean_before,
                      ))  # m[self.primary_var_ii].mean()))
        except (UnboundLocalError, NameError, ValueError, TypeError):
            pass

        if self.plot:
            ts.corr_img(sim_sea, 0,
                        "Correlation matrix of backtransformed simulated data",
                        self.var_names)

        # lets store this so we can play  with it from the outside
        self.sim, self.sim_sea = sim, sim_sea

        dump_data(self.sim_times, self.sim_sea, self.var_names, self.p,
                  0 if self.q is None else self.q,
                  random_state=pre_sim_random_state)

        return self.sim_times, sim_sea

    def disaggregate(self, var_names_dis):
        """ disaggregate variables to hourly time steps by drawing from the
        residuals of the measured data

        Parameters
        ----------
        varnames_dis : list of str
            names of variables to be disaggregated
        sim_sea : array
            simulated variables in daily time steps

        Returns
        -------
        times_out : array of datetime objects
        sim_sea_dis : array
            simulated variables in hourly time steps. the last day gets lost
        """
        #### disaggregate selected variables to hourly values, all the others
        #will be repeated 24 times a day. The last day gets lost
        print "Disaggregating selected variables."

        # how many timesteps are there per day in the input data?
        # we assume that the data is not finer than hours
        hours = np.unique([dtime.hour for dtime in self.times_orig])
        # read timesteps per day (abbreviation, because it is used a lot as an
        # index modifier below
        tpd = len(hours)

        times_out = sim_times((self.T - 1) * tpd, 1., self.start_date)[0]
        doys_out = times.datetime2doy(times_out)
        # these are in "hourly" discretization
        doys_in = times.datetime2doy(self.times_orig[:-tpd])

        # due to the interpolation with interp1d, we often have to skip the
        # last day (24 hours)
        # size of pool we can draw from:
        nn = self.T_data // tpd * tpd - tpd
        m = self.T * tpd - tpd  # no of hourly timesteps (simulated)
        sim_sea_dis = self.sim_sea.repeat(tpd).reshape(-1, m + tpd)[:, :-tpd]
        deltas_input = np.zeros((self.K, nn))
        sim_interps = np.copy(sim_sea_dis)
        for var_name in var_names_dis:
            if var_name == "Qsw":
                # qsw is treated differently (physical boundaries)
                continue
            var_i = self.var_names.index(var_name)
            # hourly measured values
            var_h = my.interp_nan(self.met[var_name][:nn + tpd])
            # daily averages of measured values:
            var_d = my.nanavg(var_h.reshape(-1, tpd), axis=1)
            # 'time' array for interp1d
            hourly_input_times = np.arange(var_h.shape[0])
            # interpolate between daily values (measurements)
            f_interp = interpolate.interp1d(hourly_input_times[::tpd], var_d)
            var_interp = f_interp(hourly_input_times[:-tpd])
            # residuals to draw from
            deltas_input[var_i] = var_h[:-tpd] - var_interp
            # if the variable has lower/upper limits, store the percentage of
            # the distance covered between the interpolated line and the limit
            limits = conf.par_known[var_name]
            if limits is not None and ("l" in limits or "lc" in limits):
                lower_func = limits["l"] if "l" in limits else limits["lc"]
                lower = lower_func(doys_in) / self.sum_interval[var_i]
                lower_perc = deltas_input[var_i] / (var_interp - lower)
                lower_perc[np.isnan(lower_perc)] = 0
                deltas_input[var_i] = np.where(deltas_input[var_i] < 0,
                                               lower_perc, deltas_input[var_i])
            if limits is not None and ("u" in limits or "uc" in limits):
                upper_func = limits["u"] if "u" in limits else limits["uc"]
                upper = upper_func(doys_in) / self.sum_interval[var_i]
                upper_perc = deltas_input[var_i] / (upper - var_interp)
                upper_perc[np.isnan(upper_perc)] = 0
                deltas_input[var_i] = np.where(deltas_input[var_i] > 0,
                                               upper_perc, deltas_input[var_i])
            # not a real 'time' as above
            hourly_output_times = np.arange(self.T * tpd)
            # interpolate between daily values (simulation)
            f_interp = interpolate.interp1d(hourly_output_times[::tpd],
                                            self.sim_sea[var_i])
            sim_interps[var_i] = f_interp(hourly_output_times[:-tpd])
        # drawing from pool: use same indices for all to preserve dependencies
        deltas_drawn = deltas_input[:, self._shuffle(nn, m, tpd, 2 * tpd)]
        sim_sea_dis = sim_interps + deltas_drawn

        # check for variables with limits
        for var_name in var_names_dis:
            if var_name == "Qsw":
                continue
            var_i = self.var_names.index(var_name)
            limits = conf.par_known[var_name]
            if limits is not None and ("l" in limits or "lc" in limits):
                lower_func = limits["l"] if "l" in limits else limits["lc"]
                lower = lower_func(doys_out) / self.sum_interval[var_i]
                lower_perc = sim_interps + deltas_drawn * (sim_interps - lower)
                sim_sea_dis[var_i] = np.where(deltas_drawn[var_i] < 0,
                                              lower_perc[var_i],
                                              sim_sea_dis[var_i])
            if limits is not None and ("u" in limits or "uc" in limits):
                upper_func = limits["u"] if "u" in limits else limits["uc"]
                upper = upper_func(doys_out) / self.sum_interval[var_i]
                upper_perc = sim_interps + deltas_drawn * (upper - sim_interps)
                sim_sea_dis[var_i] = np.where(deltas_drawn[var_i] > 0,
                                              upper_perc[var_i],
                                              sim_sea_dis[var_i])
        if "Qsw" in var_names_dis:
            times_unix = times.datetime2unix(times_out[:-(tpd - 1)])
            var_i = self.var_names.index("Qsw")
            sim_sea_dis[var_i] = \
                sw_daily(times_unix, self.sim_sea[var_i, :-1])[1]

        #### if varnames_dis==None, daily values are returned and dumped:
        dump_data(times_out, sim_sea_dis, self.var_names, self.p,
                  0 if self.q is None else self.q, "_disaggregated")

        if self.plot:
            valid_rows = [self.var_names.index(var_name)
                          for var_name in var_names_dis]
            hourly_data = met_as_array(self.met, var_names=var_names_dis)
            plot_time_hists(hourly_data, self.times_orig, "%s hourly raw data",
                       var_names_dis, all_var_names=var_names_dis)
            plot_time_hists(sim_sea_dis[valid_rows], times_out,
                       "%s disaggregated data", var_names_dis,
                       all_var_names=var_names_dis)
            ts.plot_auto_corr(hourly_data, 2 * tpd,
                              "Autocorrelation of hourly measured variables",
                              var_names=var_names_dis)
            ts.plot_auto_corr(sim_sea_dis[valid_rows], 2 * tpd,
                              "Autocorrelation of disaggregated variables",
                              var_names=var_names_dis)
            ts.corr_img(hourly_data, 0,
                        "Correlation matrix of hourly measured data",
                        var_names_dis)
            ts.corr_img(sim_sea_dis[valid_rows], 0,
                        "Correlation matrix of disaggregated variables",
                        var_names_dis)

        return times_out, sim_sea_dis

    def predict(self, dtimes=None, data_past=None, T=1, n_realizations=1):
        """Predict the next T steps given the past data.

        Parameters
        ----------
        dtimes : None or (p,) ndarray, optional
            datetimes of past timesteps, if None, the last p datetimes from
            the input data are used
        data_past : None or (K, p)) ndarray, optional
            Past on which to predict the future. p is the order of the VAR -
            process (it is stored in self.p after self.fit() is called).
            If None, data from the input is used.
        T : int, optional
            number of timesteps to predict.
        n_realizations : int, optional
            number of realizations of prediction

        Returns
        -------
        prediction : (K, T) or (K, T, n_realizations) ndarray
        """
        if self.p is None:
            self.fit()
        if dtimes is None:
            dtimes = self.times[-self.p:]
        if data_past is None:
            data_past = self.data_raw[:, -self.p:]
        past_doys = times.datetime2doy(dtimes)
        data_past_trans = \
            self._fit_seasonal(values=data_past, doys=past_doys)[0]
        prediction = models.VAR_LS_predict(data_past_trans, self.AM,
                                           self.sigma_u, T, n_realizations)
        prediction = prediction.reshape((-1, T, n_realizations))

        # calculate the output resolution in days
        first_doy, second_doy = times.datetime2doy(self.times[:2])
        delta_doy = second_doy - first_doy
        prediction_doys = past_doys[-1] + delta_doy * np.arange(1, T + 1)
        prediction_dtimes = times.doy2datetime(prediction_doys,
                                               year=dtimes[-1].year)
        prediction_doys = np.where(prediction_doys > 366,
                                   prediction_doys - 366, prediction_doys)
        prediction_sea = np.atleast_3d(np.empty_like(prediction))

        for r in range(n_realizations):
            prediction_sea[..., r] = \
                seasonal_back(self.dist_sol, prediction[..., r],
                              self.var_names, doys=prediction_doys)
        prediction_sea /= self.sum_interval[:, np.newaxis, :]
        return prediction_dtimes, np.squeeze(prediction_sea)

    def plot_prediction(self, dtimes=None, data_past=None, T=1, hindcast=True):
        # don't be verbose for this part (we fetch previous fits all the time
        # which is rather boring)
        verbose = self.verbose
        self.verbose = False
        if self.p is None:
            self.fit()
        n_prev_steps = (4 if hindcast else 2) * self.p
        if dtimes is None:
            dtimes = self.times[-n_prev_steps:]
        if data_past is None:
            data_past = self.data_raw[:, -n_prev_steps:]

        prediction_dtimes, prediction = \
            self.predict(dtimes[-self.p:], data_past[:, -self.p:], T)

        fig, axes = plt.subplots(nrows=self.K, sharex=True, squeeze=True)
        fig.suptitle("Prediction made on %s" %
                     times.datetime2str(times.datetime.now()))
        for var_i, (ax, var_name) in enumerate(zip(axes, self.var_names)):
            ax.plot(dtimes[self.p:],
                    data_past[var_i, self.p:] / self.sum_interval[var_i], "k",
                    linewidth=2)
            ax.plot(prediction_dtimes, prediction[var_i], "b-x", linewidth=2)
            ax.grid()
            ax.set_ylabel(r"%s %s" % (var_name, conf.units[var_name]))
            ax.set_title(conf.long_var_names[var_name])

        if hindcast:
            for shift in range(self.p + 1, n_prev_steps):
                hindcast_dtimes, hindcast = \
                    self.predict(dtimes[:shift], data_past[:, :shift],
                                 n_prev_steps - shift, 1)
                col = plt.cm.jet(1 -
                                 float(shift - self.p - 1) /
                                 (n_prev_steps - self.p - 1),
                                 alpha=.5)
                for var_i, ax in enumerate(axes):
                    ax.plot(hindcast_dtimes, hindcast[var_i], "-",
                            color=col, linewidth=1)
                    # connect the hindcast with the end of its supporting data
                    ax.plot([dtimes[shift - 1], hindcast_dtimes[0]],
                            [data_past[var_i, shift - 1] /
                             self.sum_interval[var_i],
                             np.atleast_1d(hindcast[var_i])[0]], "-x",
                            color=col, linewidth=1)
        ti = plt.xticks()[0]
        la_ = [times.datetime2str(times.ordinal2datetime(tii), "%d.%m. %H h")
               for tii in ti]
        plt.xticks(ti, la_, rotation=30)

        self.verbose = verbose

    def plot_doy_scatter(self, var_names=None, opacity=.4):
        """Plot variables over doys. Measurement and backtransformed simulated
        data will be plotted, if it is available.

        Parameters
        ----------
        var_names : sequence of str, optional
            Which variables to plot. None means plot all.
        opacity : float in the range (0,1], optional
            Opacity of the scatter points.

        .. plot::

            met_vg = VG(("theta",))
            sim_times, sim = met_vg.simulate()
            met_vg.plot_doy_scatter()

        """
        if var_names is None:
            var_names = self.var_names
        # when given a string, this does not separate the letters, as tuple()
        # or list does
        var_names = np.atleast_1d(var_names)
        figs = []
        for var_name in var_names:
            var_ii = self.var_names.index(var_name)
            fig = plt.figure()
            figs.append(fig)
            plt.scatter(self.data_doys,
                        self.data_raw[var_ii] / self.sum_interval[var_ii],
                        marker="o", edgecolors=(0, 0, 1, opacity),
                        facecolors=(0, 0, 0, 0), label="measured")

            if hasattr(self, "sim_sea"):
                plt.scatter(self.sim_doys, self.sim_sea[var_ii], marker="o",
                            edgecolors=(1, 0, 0, opacity),
                            facecolors=(0, 0, 0, 0),
                            label="simulated backtransformed")
            plt.xlabel("doy")
            plt.ylabel("%s %s" % (var_names_greek(self.var_names)[var_ii],
                                  conf.units[var_name]))
            plt.legend(loc="best")
            plt.grid()
            title_str = "Seasonal scatter %s" % conf.long_var_names[var_name]
            plt.xlim(0, 366)
            plt.title(title_str)
            fig.canvas.set_window_title("%s (%d)" %
                                        (title_str, fig.canvas.manager.num))
        return figs

    def plot_doy_scatter_residuals(self, var_names=None, opacity=.4):
        if var_names is None:
            var_names = self.var_names
        # when given a string, this does not separate the letters, as tuple()
        # or list does
        var_names = np.atleast_1d(var_names)
        figs = []
        for var_ii, var_name in enumerate(var_names):
            fig = plt.figure()
            figs.append(fig)
            plt.scatter(self.data_doys,
                        self.data_trans[var_ii], marker="o",
                        edgecolors=(0, 0, 1, opacity),
                        facecolors=(0, 0, 0, 0), label="transformed")

            plt.scatter(self.data_doys, self.residuals[var_ii], marker="o",
                        edgecolors=(1, 0, 0, opacity),
                        facecolors=(0, 0, 0, 0),
                        label="residuals")
            plt.xlabel("doy")
            plt.ylabel("%s %s" % (var_names_greek(self.var_names)[var_ii],
                                  conf.units[var_name]))
            plt.legend()
            plt.grid()
            title_str = ("Seasonal scatter of residuals %s" %
                         conf.long_var_names[var_name])
            plt.title(title_str)
            fig.canvas.set_window_title("%s (%d)" %
                                        (title_str, fig.canvas.manager.num))
        return figs

    def plot_spaces(self, var_names=None, opacity=.4):
        if var_names is None:
            var_names = self.var_names

        figs = []
        for var_ii, var_name1 in enumerate(var_names):
            for var_jj in range(var_ii + 1, len(var_names)):
                var_name2 = var_names[var_jj]
                fig = plt.figure()
                figs.append(fig)
                plt.scatter(self.data_trans[var_ii], self.data_trans[var_jj],
                            marker="o", edgecolors=(0, 0, 1, opacity),
                            facecolors=(0, 0, 0, 0), label="transformed data")
                plt.scatter(self.sim[var_ii], self.sim[var_jj], marker="o",
                            edgecolors=(1, 0, 0, opacity),
                            facecolors=(0, 0, 0, 0), label="simulated")
                plt.legend()
                plt.xlabel(var_name1)
                plt.ylabel(var_name2)
                plt.grid()
                title_str = ("Spaces %s over %s" %
                             (conf.long_var_names[var_name2],
                              conf.long_var_names[var_name1]))
                plt.title(title_str)
                fig.canvas.set_window_title("%s (%d)" %
                                            (title_str,
                                             fig.canvas.manager.num))
        return figs

    def plot_diff_spaces(self, var_names=None, diff=1, opacity=.4):
        if var_names is None:
            var_names = self.var_names

        figs = []
        for var_ii, var_name in enumerate(var_names):
            fig = plt.figure()
            figs.append(fig)
            plt.scatter(self.data_trans[var_ii, diff:],
                        self.data_trans[var_ii, :-diff],
                        marker="o", edgecolors=(0, 0, 1, opacity),
                        facecolors=(0, 0, 0, 0), label="transformed data")
            plt.scatter(self.sim[var_ii, diff:], self.sim[var_ii, :-diff],
                        marker="o", edgecolors=(1, 0, 0, opacity),
                        facecolors=(0, 0, 0, 0), label="simulated")
            plt.legend()
            plt.xlabel("t")
            plt.ylabel("t - %d" % diff)
            plt.grid()
            title_str = ("Spaces %d-lagged %s" %
                         (diff, conf.long_var_names[var_name]))
            plt.title(title_str)
            fig.canvas.set_window_title("%s (%d)" %
                                        (title_str,
                                         fig.canvas.manager.num))
        return figs

    def plot_timehists(self, data=None, titles=None, times=None,
                       var_names=None, *args, **kwds):
        """Plots time series together with histograms.

        Parameters
        ----------
        data : 2dim ndarray or sequence of 2dim ndarrays, optional
            Which data type to plot. If none, data_raw is used.
        titles : sequence of str, optional
        var_names : sequence of str, optional
            Which variables to plot. None means plot all.
        """
        # TODO: this is ugly to use
        if data is None:
            data = self.data_raw / self.sum_interval
        if times is None:
            times = self.times
        if titles is None:
            titles = len(data) * ["%s"]
        if np.rank(data) == 2:
            data = data[np.newaxis, ...]
        if var_names is None:
            var_names = self.var_names
        figs = []
        for data_2d, title in zip(data, titles):
            figs += [plot_time_hists(data_2d,
                                     times, title, var_names, self.ylabels,
                                     self.var_names, *args, **kwds)]
        return figs

    def plot_seasonal_fit(self, var_names=None, plot_quantiles=True,
                            plot_fourier=True, opacity=.25, n_bins=15,
                            kde=True, *args, **kwds):
        """Plots seasonal scatter with fitted densities and a histogram of the
        transformed quantiles.

        Parameters
        ----------
        var_names : sequence of str, optional
            Which variables to plot. None means plot all.
        plot_quantiles : boolean, optional
            Plot histograms of the quantiles?
        """
        if var_names is None:
            var_names = self.var_names
        elif type(var_names) is str:
            var_names = (var_names,)

        figs = []
        for var_ii, var_name in enumerate(var_names):
            dist, solution = self.dist_sol[var_name]
            if var_name in conf.long_var_names:
                long_var_name = conf.long_var_names[var_name]
            else:
                long_var_name = var_name
            fig = dist.scatter_pdf(solution, title=long_var_name,
                                   n_sumup=self.sum_interval.ravel()[var_ii],
                                   opacity=opacity, *args, **kwds)
            plt.ylabel(conf.units[var_name]
                       if var_name in conf.units else "[?]")
            fig.canvas.set_window_title("Scatter pdf %s (%d)"
                                        % (var_name,
                                           fig.canvas.manager.num))
            figs += [fig]

            if plot_quantiles:
                fig = plt.figure(*args, **kwds)
                var_ii = self.var_names.index(var_name)
                quantiles = dist.cdf(solution, x=self.data_raw[var_ii])
                my.hist(quantiles, n_bins, fig=fig, kde=kde)
                plt.title("Quantiles %s" % var_name)
                fig.canvas.set_window_title("Quantiles %s (%d)"
                                            % (var_name,
                                               fig.canvas.manager.num))
                figs += [fig]

            if plot_fourier:
                if isinstance(dist, sd.SeasonalDist):
                    fig = dist.plot_fourier_fit()
                    fig.suptitle(var_name)
                    figs += [fig]
        return figs

    def plot_autocorr(self, maxlag=7, *args, **kwds):
        """Plots autocorrelation of transformed and simulated data (if the
        latter is available).

        Parameters
        ----------
        maxlag : int, optional
            Maximum number of lags to plot
        """
        data = self.data_trans
        figs = []
        if self.sim is not None:
            data = np.array([data, self.sim])
        figs += [ts.plot_auto_corr(data, maxlag,
                   "Autocorrelation of standard-normal distributed variables",
                   var_names_greek(self.var_names), *args, **kwds)]

        data = self.data_raw
        if self.sim_sea is not None:
            data = np.array([data, self.sim_sea])
        figs += [ts.plot_auto_corr(data, maxlag,
                   "Autocorrelation of untransformed variables",
                   var_names_greek(self.var_names), *args, **kwds)]
        figs += [ts.plot_auto_corr(self.residuals, 7,
                                   "Autocorrelation of the residuals.",
                                   self.var_names)]
        return figs

    def _shuffle(self, nn, m, tpd=24, autocorr_len=48):
        """Draw a clustered sample of size m with elements [0,nn]."""

        # like a normal modulus, but knows how to divide by 0!!
        mod0 = lambda x, y: x % y if y != 0 else 0

        def implant_chunk(ii):
            dst_point = np.random.randint(m)
            # hour of day from 0 to 23 in destination
            hour_of_dst = mod0(dst_point, tpd)
            src_point = np.random.randint(nn)
            # insure the same hour of day in source
            src_point += -mod0(src_point, tpd) + hour_of_dst
            src_point = min(src_point, nn - 1)
            dst_len = len(ii[dst_point:dst_point + autocorr_len])
            chunk = np.arange(src_point, src_point + dst_len)
            chunk[chunk >= (nn - 1)] = nn - 1
            ii[dst_point:dst_point + autocorr_len] = chunk
            return ii

        # indices has to be of length m with elements no bigger than nn
        indices = np.arange(m) % nn

        # i pulled the number of chunks out of my ass, there is no theory to
        # it at this point.
        # with this being said, shake it baby!
        for how_long_do_i_have_to_stay in xrange(2 * int(m / autocorr_len)):
            indices = implant_chunk(indices)

        return indices

    def _load_and_prepare_data(self, delimiter="\t", **met_kwds):
        """Loading the data from the met_file and aggregate it according to
        sum_interval (both defined in __init__). Plus plotting if requested."""
        # get the data
        # this covers the case that a non-default met filename was given that
        # exists in the data_dir but not in the working path
        if (self.met_file is not None) and (not os.path.exists(self.met_file)):
            self.met_file = os.path.join(self.data_dir, self.met_file)

        if not os.path.exists(self.cache_dir):
            os.mkdir(self.cache_dir)

        self.times_orig, self.met = \
            read_met(self.met_file, delimiter=delimiter, verbose=self.verbose,
                     **met_kwds)

        # convert the dictionary into an array.
        # mind the order: alpha-numeric according to var_names
        # will be shorter once summed up
        T_raw_data = len(self.met[self.var_names[0]])
        # self.K corresponds to all variables. the wtemp variables are
        # unfortunately stored someplace else and should be ignored here
        # so for the shape of data, we do not count to K, but something lower
        # if there are water temperatures in var_names
        var_names_part = [var_name for var_name in self.var_names
                          if (not var_name.startswith("wtemp")
                              and var_name != "nao")]
        data = np.nan * np.empty((len(var_names_part),
                                  T_raw_data / self.sum_interval[0, 0]))
        for key in var_names_part:
            var_i = var_names_part.index(key)
            if self.sum_interval[var_i] > 1:
                data[var_i], times_ = my.sumup(self.met[key],
                                               self.sum_interval[var_i],
                                               self.times_orig,
                                               middle_time=False)
            else:
                data[var_i] = self.met[key]
                times_ = self.times_orig

# TODO: if the following blocks work is a matter of order in var_names. Ugly!

        if "U" in self.var_names:
            # the wind has to be aggregated taking the direction into account
            data[self.var_names.index("U")] = \
                daily_wind(self.met, self.sum_interval_dict["U"])[1]

#        if "nao" in self.var_names:
#            import nao
#            nao_dates, naoi = nao.read_nao()
#            self.met["nao"] = naoi[np.where(nao_dates == times_[0])[0]:
#                              np.where(nao_dates == times_[-1])[0] + 1]
#            var_row = self.var_names.index("nao")
#            summed_nao = my.sumup(self.met["nao"], self.sum_interval[var_row])
#            if summed_nao.shape[1] < data.shape[1]:
#                shape_diff = data.shape[1] - summed_nao.shape[1]
#                summed_nao = np.concatenate((np.zeros((1, shape_diff)),
#                                             summed_nao), axis=1)
#            data = np.insert(data, var_row, summed_nao, axis=0)

        # check if we are supposed to simulate water temperatures - they are
        # stored someplace else
#        if "wtemp" in "".join(self.var_names):
#            met_rivers = read_met(conf.hyd_file)[1]
#            self.met.update(met_rivers)
#            for var_name in met_rivers.keys():
#                if var_name in self.var_names:
#                    var_row = self.var_names.index(var_name)
#                    new_data = my.sumup(self.met[var_name],
#                                        self.sum_interval[var_row])
#                    # the last measurement of data was dropped in sumup! so
#                    # loose it also here
#                    new_data = new_data[:, :data.shape[1]]
#                    data = np.insert(data, var_row, new_data, axis=0)
        return times_, data

    def _prepare_fixed_data(self):
        """Convert the data to standard-normal an put it into (K,T)-array form.
        """
        if self.fixed_variables:
            sh = shelve.open(self.seasonal_cache_file, "c")
            fixed_data = np.nan * np.empty((self.K, self.T))
            for var_name, values in self.fixed_variables.items():
                var_ii = self.var_names.index(var_name)
                if values is None:
                    # fix the input data the model was fitted on
                    values = self.data_raw[var_ii]
                solution_key = var_name

                # the fitting was originially done for daily sums of hourly
                # values, but we can have different aggregation lengths
                sum_interval = self.sum_interval_dict[var_name]
                if sum_interval != 24:
                    solution_key += "_%d" % sum_interval
                seas_class, dist_class, solution = sh[solution_key]
                dist = seas_class(dist_class, values, self.times,
                                  fixed_pars=conf.par_known[var_name])
                quantiles = np.squeeze(dist.cdf(solution))
                transformed = distributions.norm.ppf(quantiles)
                fixed_data[var_ii] = transformed
            sh.close()
            return fixed_data
        else:
            return None

    def _fit_seasonal(self, refit=None, values=None, doys=None):
        if refit is None:
            refit = tuple()
        elif refit == "all" or refit:
            refit = self.var_names
        sh = shelve.open(self.seasonal_cache_file, 'c')
        if values is None:
            values = self.data_raw
        if doys is None:
            doys = self.data_doys
        data_trans = np.empty_like(values)
        dist_sol = {}
        for var_ii, var in enumerate(values):
            var_name = self.var_names[var_ii]
            solution_key = var_name

            # the fitting was originially done for daily sums of hourly values
            # we can have different aggregation lengths so we trigger a fitting
            # if we come across one of these
            sum_interval = self.sum_interval_dict[var_name]
            plain_solution_key = solution_key
            if sum_interval != 24:
                solution_key += "_%d" % sum_interval

            if solution_key not in sh.keys() or plain_solution_key in refit:
                if self.verbose:
                    print "\tFitting a distribution to ", var_name
                seas_class = conf.seasonal_classes[var_name]
                if seas_class is skde.SeasonalKDE:
                    dist = seas_class(var, self.times,
                                      fixed_pars=conf.par_known[var_name],
                                      verbose=self.verbose)
                    solution = dist.fit()
                    sh[solution_key] = [seas_class, None, solution]
                else:
                    dist = seas_class(conf.dists[var_name], var, self.times,
                                      fixed_pars=conf.par_known[var_name],
                                      verbose=self.verbose)
                    solution = dist.fit()
                    sh[solution_key] = \
                        [seas_class, conf.dists[var_name], solution]
            else:
                if self.verbose:
                    print "\tRecover previous fit from shelve for: ", var_name
                seas_class, dist_class, solution = sh[solution_key]
                if len(self.times) > len(var):
                    # we assume we were called from predict in which case
                    # we have to pass the past data in here and not the
                    # passed data, hehe
                    var_ = self.data_raw[var_ii]
                else:
                    var_ = var
                if seas_class is skde.SeasonalKDE:
                    dist = seas_class(var_, self.times, solution,
                                      fixed_pars=conf.par_known[var_name],
                                      verbose=self.verbose)
                else:
                    dist = seas_class(dist_class, var_, self.times,
                                      solution=solution,
                                      fixed_pars=conf.par_known[var_name])

            quantiles = dist.cdf(solution, x=var, doys=doys)
            data_trans[var_ii] = distributions.norm.ppf(quantiles)
            dist_sol[var_name] = (dist, solution)
        sh.close()
        return data_trans, dist_sol

    def _scenario_parameters(self, sim_doys):
        """Prepares the scenario parameters, S, m and m_trend for the
        std.-normal world."""
        m = S = m_t = m_trend = None
        # scale other variables according to linear regression
        scale = (np.cov(self.data_trans)[self.primary_var_ii] /
                 np.var(self.data_trans[self.primary_var_ii]))
        intercept = (self.data_trans.mean(axis=1) -
                     scale * self.data_trans[self.primary_var_ii].mean()
                     ).reshape((self.K, 1))
#        scale = (np.cov(self.data_raw)[self.primary_var_ii] /
#                 np.var(self.data_raw[self.primary_var_ii]))
#        intercept = (self.data_raw.mean(axis=1) -
#                     scale * self.data_raw[self.primary_var_ii].mean()
#                     ).reshape((self.K, 1))
        # do not change the temperature further than wanted
        scale[self.primary_var_ii] = 1
        intercept[self.primary_var_ii] = 0
        if self.primary_var in self.var_names:
            seas_dist, trig_params = self.dist_sol[self.primary_var]
            # get the doy-specific distribution parameters
            _T = (2 * np.pi / 365 * sim_doys)[np.newaxis, :]
            sigma = seas_dist.trig2pars(trig_params, _T)[1]
            # e.g. gradual temperature change
            primvar_trend = \
                self.theta_grad * self.sum_interval[self.primary_var_ii]
            # looking for the mean of the standard deviation
            m_trend_primvar = primvar_trend / np.mean(sigma)
            m_trend = m_trend_primvar * scale

            # change in mean in the real world
            delta_primvar = \
                self.theta_incr * self.sum_interval[self.primary_var_ii]
            m_primvar = delta_primvar / sigma
            m = m_primvar * scale.reshape((self.K, 1)) + intercept
#            target_means = ((delta_primvar +
#                             self.data_raw[self.primary_var_ii].mean())
#                            * scale + intercept.ravel())
#            delta_medians = target_means - np.mean(self.data_raw, axis=1)
#            # get the corresponding value in the standard-normal world by
#            # qq-transform
#            sh = shelve.open(self.seasonal_cache_file, 'r')
#            m = []
#            for var_ii, var_name in enumerate(self.var_names):
#                solution_key = var_name
#                sum_interval = self.sum_interval_dict[var_name]
#                if sum_interval != 24:
#                    solution_key += "_%d" % sum_interval
#                seas_class, dist_class, solution = sh[solution_key]
#                if seas_class is skde.SeasonalKDE:
#                    dist = seas_class(self.data_raw[var_ii], self.times,
#                                      solution,
#                                      fixed_pars=conf.par_known[var_name])
#                else:
#                    dist = seas_class(dist_class, self.data_raw[var_ii],
#                                      self.times,
#                                      fixed_pars=conf.par_known[var_name])
#
#                # get the medians, add the delta means and calculate the std-
#                # normal equivalent
#                medians = dist.ppf(solution, quantiles=[.5] * self.T,
#                                   doys=sim_doys)
#                quantiles = dist.cdf(solution,
#                                     x=medians + delta_medians[var_ii],
#                                     doys=sim_doys)
#                m += [distributions.norm.ppf(quantiles)]
#            sh.close()

            S = np.asmatrix(np.identity(self.K, dtype=float))

            # das war an der Tafel im Seminarraum 2:
            # mean_arrival: mittlere Zeit zwischen Aenderungen (kalte/warme
            # Perioden)
            # disturbance_std: Stabw der disturbances
            disturbance_std = self.disturbance_std * \
                        self.sum_interval[self.primary_var_ii] / np.mean(sigma)
            m_t_primvar = interannual_variability(self.T, self.mean_arrival,
                                                disturbance_std)
            m_t = m_t_primvar * scale.reshape((self.K, 1))

            if self.climate_signal is not None:
                climate_diff = self._get_climate_diff(_T) / sigma
                m_t += climate_diff * scale.reshape((self.K, 1))

        return S, m, m_t, m_trend

    @property
    def fitted_theta_means(self):
        if not hasattr(self, "_T"):
            # get the doy-specific distribution parameters
            _T = (2 * np.pi / 365 * self.sim_doys)[np.newaxis, :]
        else:
            _T = self._T
        dist, solution = self.dist_sol[self.primary_var]
        return (dist.trig2pars(solution, _T)[0] /
                self.sum_interval[self.primary_var_ii])

    def _get_climate_diff(self, _T):
        """Calculates the difference between the seasonal mean of the primary
        variable and the given climate signal."""
        return ((self.climate_signal - self.fitted_theta_means) *
                self.sum_interval_dict[self.primary_var])

if __name__ == "__main__":
    met_vg = VG(("theta", "ILWR", "Qsw", "rh", "u", "v"), verbose=True)
    met_vg.fit(3)
    st, sim = met_vg.simulate()
    dst, dsim = met_vg.disaggregate(("Qsw", "u", "v"))
    met_vg.plot_doy_scatter()
    plt.show()
