import warnings
from abc import ABCMeta, abstractmethod
import numpy as np
import matplotlib.pyplot as plt
from scipy import stats, special
import scipy.optimize as sp_optimize
try:
    import numexpr as ne
    NE = True
except ImportError:
    NE = False

# some special functions vectorized to be able to handle arrays as input
gamma_func = np.vectorize(lambda x: special.gamma(x))
gammaln = np.vectorize(lambda x: special.gammaln(x))
gammainc = np.vectorize(lambda a, x: special.gammainc(a, x))
gammaincinv = np.vectorize(lambda a, q: special.gammaincinv(a, q))
digamma = np.vectorize(lambda a: special.digamma(a))
hyp1f1 = np.vectorize(lambda a, b, x: special.hyp1f1(a, b, x))
nctdtr = np.vectorize(lambda x1, x2, x3: special.nctdtr(x1, x2, x3))
nctdtrit = np.vectorize(lambda x1, x2, x3: special.nctdtrit(x1, x2, x3))
stdtr = np.vectorize(lambda df, x: special.stdtr(df, x))
stdtrit = np.vectorize(lambda df, q: special.stdtrit(df, q))


def max_likelihood(density_func, x0, values=None, opt_func=None, disp=False,
                   *args, **kwds):
    """Fits parameters of a density function to data, using the log-maximum-
    likelihood approach."""
    if opt_func is None:
        from scipy import optimize
        opt_func = optimize.fmin
    # i hereby define the unlikelihood function as the negative log-likelihood
    # function, so that i can maximize the likelihood by minimizing the
    # unlikelihood
    if values is None:
        args = ()
        unlikelihood = lambda params:-np.sum(np.log(density_func(params)))
    else:
        args = (values,)
        def unlikelihood(params, values):
            # we have to express fixed values in terms of kwds, because there
            # could be collisions otherwise...

            # first come, first serve: interpret the first n elements of params
            # as the first n variables in dist.parameter_names
            p_kwds = {par_name: value for par_name, value in
                      zip(density_func.im_self.parameter_names, params)
                      if par_name not in kwds}
            p_kwds.update(kwds)
            return -np.sum(np.log(density_func(values, **p_kwds)))

    return opt_func(unlikelihood, x0, args=args, maxiter=1e4 * len(x0),
                    maxfun=1e4 * len(x0), disp=disp)


def min_ks(cdf, x0, values, opt_func=None, disp=False, *args, **kwds):
    """Fits parameters of a cumulative distribution function to data,
    minimizing the Kolmogorov-Smirnof test statistic."""
    if opt_func is None:
        opt_func = sp_optimize.fmin
    n = len(values)
    values_sorted = np.sort(values)
    ranks_plus = np.arange(0., n) / n
    ranks_minus = np.arange(1., n + 1) / n

    def ks(params):
        # we have to express fixed values in terms of kwds, because there
        # could be collisions otherwise...

        # first come, first serve: interpret the first n elements of params
        # as the first n variables in dist.parameter_names
        p_kwds = {par_name: value for par_name, value in
                  zip(cdf.im_self.parameter_names, params)
                  if par_name not in kwds}
        p_kwds.update(kwds)

        cdf_values = cdf(values_sorted, **p_kwds)
        cdf_values[np.isnan(cdf_values)] = np.inf
        dmin_plus = np.abs(cdf_values - ranks_plus).max()
        dmin_minus = np.abs(cdf_values - ranks_minus).max()
        return max(dmin_plus, dmin_minus)

    return opt_func(ks, x0, args=args, maxiter=1e4 * len(x0),
                    maxfun=1e4 * len(x0), disp=disp)


def min_fsum(cdf, x0, values, opt_func=None, disp=False, *args, **kwds):
    if opt_func is None:
        opt_func = sp_optimize.fmin
    n = len(values)
    values_sorted = np.sort(values)
    ranks_plus = np.arange(0., n) / n
    ranks_minus = np.arange(1., n + 1) / n

    def ks(params):
        # we have to express fixed values in terms of kwds, because there
        # could be collisions otherwise...

        # first come, first serve: interpret the first n elements of params
        # as the first n variables in dist.parameter_names
        p_kwds = {par_name: value for par_name, value in
                  zip(cdf.im_self.parameter_names, params)
                  if par_name not in kwds}
        p_kwds.update(kwds)

        cdf_values = cdf(values_sorted, **p_kwds)
        cdf_values[np.isnan(cdf_values)] = np.inf
        dmin_plus = ((cdf_values - ranks_plus) ** 2).sum()
        dmin_minus = ((cdf_values - ranks_minus) ** 2).sum()
        return dmin_plus + dmin_minus

    return opt_func(ks, x0, args=args, maxiter=1e4 * len(x0),
                    maxfun=1e4 * len(x0), disp=disp)


"""Let's mimic the the part of the interface of stats.rv_continuous that we use
in this script to implement more-than-two-parametric distributions and not
change our other code."""


class MyDist(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def pdf(self):
        pass

    @abstractmethod
    def cdf(self):
        pass

    @abstractmethod
    def ppf(self):
        pass

    @abstractmethod
    def fit(self):
        pass

    def __call__(self, *params):
        return Frozen(self, *params)

    def fit_ml(self, values, opt_func=sp_optimize.fmin, x0=None, *args,
               **kwds):
        if x0 is None:
            x0 = (1,) * len(self.parameter_names)
        return max_likelihood(self.pdf, x0, values, opt_func, *args, **kwds)

    def fit_ks(self, values, opt_func=sp_optimize.fmin, x0=None, *args,
               **kwds):
        if x0 is None:
            x0 = (1,) * len(self.parameter_names)
        return min_ks(self.cdf, x0, values, opt_func, *args, **kwds)

    def fit_fsum(self, values, opt_func=sp_optimize.fmin, x0=None, *args,
                 **kwds):
        if x0 is None:
            x0 = (1,) * len(self.parameter_names)
        return min_fsum(self.cdf, x0, values, opt_func, *args, **kwds)

    def _constraints(self, *args, **kwds):
        return True


class Frozen(object):
    # this idea is mainly stolen from the scipy.stats-module
    def __init__(self, dist, *params):
        self.dist = dist
        self.params = params

    def pdf(self, x):
        return self.dist.pdf(x, *self.params)

    def cdf(self, y):
        return self.dist.cdf(y, *self.params)

    def ppf(self, x):
        return self.dist.ppf(x, *self.params)

    def plot_fit(self, values, n_classes=30):
        """Display a combined plot of a histogram, fitted pdf, empirical cdf
        and fitted cdf."""
        fig = plt.figure()
        ax1 = fig.add_subplot(111)

        # the histogram of the data
        bins = ax1.hist(values, n_classes, normed=True, facecolor='green',
                        alpha=0.75)[1]

        class_middles = 0.5 * (bins[1:] + bins[:-1])
        density = self.pdf(class_middles)
        ax1.plot(class_middles, density, 'r--')

        # the quantile part
        ax2 = ax1.twinx()
        # empirical cdf
        values_sort = np.sort(values)
        ranks_emp = (.5 + np.arange(len(values))) / len(values)
        ax2.plot(values_sort, ranks_emp)
        # theoretical cdf
        xx = np.linspace(values.min(), values.max(), 5e2)
        ranks_theory = self.cdf(xx)
        ax2.plot(xx, ranks_theory, 'r--')
        ax2.grid()
        plt.title(", ".join(" %s: %.3f" % (par_name, par)
                            for par_name, par
                            in zip(self.dist.parameter_names, self.params)))
        return fig

    def plot_qq(self, values, *args, **kwds):
        """A qq-plot. Scatters theoretic over empirical ranks."""
        ranks_emp = (stats.stats.rankdata(values) - .5) / len(values)
        ranks_the = self.cdf(values)
        fig = plt.figure()
        plt.scatter(ranks_emp, ranks_the, marker="o", *args, **kwds)
        plt.plot([0, 1], [0, 1])
        plt.xlim(0, 1)
        plt.ylim(0, 1)
        plt.xlabel("Empirical ranks")
        plt.ylabel("Fitted ranks")
        plt.grid()
        return fig


class Censored(MyDist):
    def __init__(self, distribution, lc= -np.inf, uc=np.inf):
        self.dist = distribution
        self.lc = lc
        self.uc = uc
        # read "lc" as "lower cut", "uc" as "upper cut"
        # mind you: these are different than lower and upper bound in i.e. the
        # beta distribution. we can also truncate the beta distribution and
        # have upper bounds above the upper cut.
        if hasattr(distribution, "parameter_names"):
            self.parameter_names = tuple(list(distribution.parameter_names) +
                                         ["lc", "uc"])

    def pdf(self, x, *args, **kwds):
        x = np.atleast_1d(x)
        args = [np.asarray(arg) for arg in args]
        lc = self.lc if "lc" not in kwds else kwds.pop("lc")
        uc = self.uc if "uc" not in kwds else kwds.pop("uc")
        density = self.dist.pdf(x, *args, **kwds)
        quantiles = self.dist.cdf(x, *args, **kwds)
        lc_above_ii = np.atleast_1d(x > lc)
        if np.any(lc_above_ii):
            lc_index = np.argmin(x[lc_above_ii])
            density[lc_index] = quantiles[lc_index]
        uc_below_ii = np.atleast_1d(x < uc)
        if np.any(uc_below_ii):
            uc_index = np.argmax(x[uc_below_ii])
            density[uc_index] = 1 - quantiles[uc_index]
        density[(x < lc) | (x > uc)] = 0
        return density

    def cdf(self, x, *args, **kwds):
        x = np.asarray(x)
        args = [np.asarray(arg) for arg in args]
        lc = self.lc if "lc" not in kwds else kwds.pop("lc")
        uc = self.uc if "uc" not in kwds else kwds.pop("uc")
        quantiles = np.atleast_1d(self.dist.cdf(x, *args, **kwds))
        below_lc_ii = np.atleast_1d(x <= lc)
        if np.any(below_lc_ii):
            quantiles[below_lc_ii] = \
                self.dist.cdf(lc, *args, **kwds)[below_lc_ii]
        above_uc_ii = np.atleast_1d(x >= uc)
        if np.any(above_uc_ii):
            quantiles[above_uc_ii] = \
                1 - self.dist.cdf(uc, *args, **kwds)[above_uc_ii]
        quantiles[quantiles < 0] = 0
        quantiles[quantiles > 1] = 1
        return np.squeeze(quantiles)

    def ppf(self, quantiles, *args, **kwds):
        quantiles = np.asarray(quantiles)
        args = [np.asarray(arg) for arg in args]
        lc = self.lc if "lc" not in kwds else kwds.pop("lc")
        uc = self.uc if "uc" not in kwds else kwds.pop("uc")
        x = self.dist.ppf(quantiles, *args, **kwds)
        lower_ii = x < lc
        if np.any(lower_ii):
            x[lower_ii] = lc[lower_ii]
        upper_ii = x > uc
        if np.any(upper_ii):
            x[upper_ii] = uc[upper_ii]
        return x

    def fit(self, x, *args, **kwds):
        try:
            x0 = self.dist._feasible_start
        except AttributeError:
            x0 = self.dist.fit(x, *args, **kwds)
        # using the cdf to fit, because of the insanity inherent in the pdf
        # (i.e. no maximum likelihood)
        return min_ks(self.cdf, x0, x, lc=self.lc, uc=self.uc, *args)

    def constraints(self, x, *args, **kwds):
        return self.dist.constraints(x, *args, **kwds)


class Truncated(MyDist):
    """Wraps around any other distribution found here to truncate its upper
    and/or lower tail."""
    def __init__(self, distribution, lc= -np.inf, uc=np.inf):
        self.dist = distribution
        self.lc = lc
        self.uc = uc
        # read "lc" as "lower cut", "uc" as "upper cut"
        # mind you: these are different than lower and upper bound in i.e. the
        # beta distribution. we can also truncate the beta distribution and
        # have upper bounds above the upper cut.
        self.parameter_names = tuple(list(distribution.parameter_names) +
                                     ["lc", "uc"])

    def pdf(self, x, *args, **kwds):
        args = [np.asarray(arg) for arg in args]
        if "uc" in kwds:
            uc = kwds.pop("uc")
            if len(args) > 0:
                del args[-1]
        else:
            uc = self.uc
        if "lc" in kwds:
            lc = kwds.pop("lc")
            if len(args) > 0:
                del args[-1]
        else:
            lc = self.lc
        if len(args) > len(self.dist.parameter_names):
            args = args[:len(self.dist.parameter_names)]
        # still sober
        un_trunc = (self.dist.pdf(x, *args, **kwds) /
                    (self.dist.cdf(uc, *args, **kwds) -
                     self.dist.cdf(lc, *args, **kwds)))
        un_trunc[(x < lc) | (x > uc)] = 1e-9
        # now totally drunkated
        return un_trunc

    def cdf(self, x, *args, **kwds):
        args = [np.asarray(arg) for arg in args]
        if "uc" in kwds:
            uc = kwds.pop("uc")
            if len(args) > 0:
                del args[-1]
        else:
            uc = self.uc
        if "lc" in kwds:
            lc = kwds.pop("lc")
            if len(args) > 0:
                del args[-1]
        else:
            lc = self.lc
        if len(args) > len(self.dist.parameter_names):
            args = args[:len(self.dist.parameter_names)]
        quantiles = ((self.dist.cdf(x, *args, **kwds) -
                      self.dist.cdf(lc, *args, **kwds)) /
                     (self.dist.cdf(uc, *args, **kwds) -
                      self.dist.cdf(lc, *args, **kwds)))
        quantiles[quantiles < 0] = 0
        quantiles[quantiles > 1] = 1
        return quantiles

    def ppf(self, quantiles, *args, **kwds):
        args = [np.asarray(arg) for arg in args]
        if "uc" in kwds:
            uc = kwds.pop("uc")
            if len(args) > 0:
                del args[-1]
        else:
            uc = self.uc
        if "lc" in kwds:
            lc = kwds.pop("lc")
            if len(args) > 0:
                del args[-1]
        else:
            lc = self.lc
        if len(args) > len(self.dist.parameter_names):
            args = args[:len(self.dist.parameter_names)]
        x = self.dist.ppf(quantiles * self.dist.cdf(uc, *args, **kwds) +
                          (1 - quantiles) * self.dist.cdf(lc, *args, **kwds),
                          *args, **kwds)
        lower_ii = x < lc
        if np.sum(lower_ii) > 0:
            x[lower_ii] = lc[lower_ii]
        upper_ii = x > uc
        if np.sum(upper_ii) > 1:
            x[upper_ii] = uc[upper_ii]
        return x

    def fit(self, x, *args, **kwds):
        try:
            x0 = self.dist._feasible_start
        except AttributeError:
            x0 = self.dist.fit(x, *args)
        # using the cdf to fit, because of the insanity inherent in the pdf
        # (i.e. no maximum likelihood)
        if self.lc is None:
            x0 = tuple(list(x0) + [np.min(x)])
            kwds["lc"] = x0[-1]
        if self.uc is None:
            x0 = tuple(list(x0) + [np.max(x)])
            kwds["uc"] = x0[-1]
        solution = list(min_ks(self.cdf, x0, x, *args, **kwds))
        if self.uc is None:
            self.uc = solution.pop(-1)
        if self.lc is None:
            self.lc = solution.pop(-1)
        if "lc" in kwds:
            solution += [kwds["lc"]]
        if "uc" in kwds:
            solution += [kwds["uc"]]
        return solution

    def constraints(self, x, *args, **kwds):
        return self.dist.constraints(x, *args, **kwds)


class Normal(MyDist):
    def pdf(self, x, mu=0, sigma=1):
        x, mu, sigma = np.asarray(x), np.asarray(mu), np.asarray(sigma)
        return (1. / (2 * np.pi * sigma ** 2) ** .5 *
                np.exp(-(x - mu) ** 2 / (2 * sigma ** 2)))

    def cdf(self, x, mu=0, sigma=1):
        return .5 * (1 + special.erf((x - mu) / (sigma * 2 ** .5)))

    def ppf(self, quantiles, mu=0, sigma=1):
        return special.ndtri(quantiles) * sigma + mu

    def fit(self, x, mu=None, sigma=None):
        x_noninf = x[np.isfinite(x)]
        if mu is None:
            mu = x_noninf.mean()
        if sigma is None:
            sigma = x_noninf.std()
        return mu, sigma
norm = Normal()
norm.parameter_names = ("mu", "sigma")
norm._feasible_start = (0, 2)


class TruncNormal(MyDist):
    def pdf(self, x, mu=0, sigma=1, lc= -np.inf, uc=np.inf):
        x, mu, sigma, lc, uc = map(np.atleast_1d, (x, mu, sigma, lc, uc))
        un_trunc = (norm.pdf(x, mu, sigma) /
                    (norm.cdf(uc, mu, sigma) - norm.cdf(lc, mu, sigma)))
        un_trunc[(x < lc) | (x > uc)] = 0
        return np.asscalar(un_trunc) if len(un_trunc) == 1 else un_trunc

    def cdf(self, x, mu=0, sigma=1, lc= -np.inf, uc=np.inf):
        x, mu, sigma, lc, uc = map(np.atleast_1d, (x, mu, sigma, lc, uc))
        q = ((norm.cdf(x, mu, sigma) - norm.cdf(lc, mu, sigma)) /
             (norm.cdf(uc, mu, sigma) - norm.cdf(lc, mu, sigma)))
        q[q < 0] = 0
        q[q > 1] = 1
        return np.asscalar(q) if len(q) == 1 else q

    def ppf(self, quantiles, mu=0, sigma=1, lc= -np.inf, uc=np.inf):
        quantiles, mu, sigma, lc, uc = map(np.atleast_1d,
                                           (quantiles, mu, sigma, lc, uc))
        x = norm.ppf(quantiles * norm.cdf(uc, mu, sigma) +
                     (1 - quantiles) * norm.cdf(lc, mu, sigma), mu, sigma)
        lower_ii = x < lc
        if np.sum(lower_ii) > 0:
            x[lower_ii] = lc[lower_ii]
        upper_ii = x > uc
        if np.sum(upper_ii) > 1:
            x[upper_ii] = uc[upper_ii]
        return np.asscalar(x) if len(x) == 1 else x

    def fit(self, x, lc= -np.inf, uc=np.inf):
        x0 = [x.mean(), x.std()]
        kwds = {}
        if lc is None:  # or np.isneginf(lc):
            x0 += [x.min()]
        else:
            kwds["lc"] = lc
        if uc is None:  # or np.isinf(uc):
            x0 += [x.max()]
        else:
            kwds["uc"] = uc
        return self.fit_ml(x, x0=x0, **kwds)
#        return self.fit_ml(x, sp_optimize.fmin, l, u)
#        xmean = x.mean()
#        xstd = x.std()
#        fac1 = ((norm.pdf(l, xmean, xstd) - norm.pdf(u, xmean, xstd)) /
#                (norm.cdf(l, xmean, xstd) - norm.cdf(u, xmean, xstd)))
#        fac2 = (((l - xmean) / xstd * norm.pdf(l, xmean, xstd) -
#                 (u - xmean) / xstd * norm.pdf(u, xmean, xstd)) /
#                (norm.cdf(l, xmean, xstd) - norm.cdf(u, xmean, xstd)))
#        mu = xmean + fac1 * xstd
#        sigma = xstd ** 2 * (1 + fac2 - fac1 ** 2)
#        return mu, sigma, l, u

    def constraints(self, x, mu, sigma, lc= -np.inf, uc=np.inf):
        if np.any(x > uc):
            return False
        if np.any(x < lc):
            return False
        return True
truncnorm = TruncNormal()
truncnorm.parameter_names = ("mu", "sigma", "lc", "uc")
truncnorm._feasible_start = (0, 1, -1, 1)


class LogNormal(MyDist):
    def pdf(self, x, mu, sigma):
        return norm.pdf(np.log(x + 1e-9), mu, sigma)

    def cdf(self, x, mu, sigma):
        return norm.cdf(np.log(x + 1e-9), mu, sigma)

    def ppf(self, quantiles, mu, sigma):
        return np.exp(norm.ppf(quantiles, mu, sigma))

    def fit(self, x):
        x_noninf = x[(x > 0) & np.isfinite(x)]
        mu = np.mean(np.log(x_noninf))
        sigma = np.std(np.log(x_noninf))
        return mu, sigma

    def _constraints(self, x, mu, sigma):
        if np.any(x <= 0):
            return False
        if np.any(sigma <= 0):
            return False
        return True
lognormal = LogNormal()
lognormal.parameter_names = ("mu", "sigma")
lognormal._feasible_start = (0, 1)


class Cauchy(MyDist):
    def pdf(self, x, x0, gamma):
        return 1 / (np.pi * gamma * (1 + ((x - x0) / gamma) ** 2))

    def cdf(self, x, x0, gamma):
        return (1 / np.pi) * np.arctan2(x - x0, gamma) + .5

    def ppf(self, q, x0, gamma):
        return x0 + gamma * np.tan(np.pi * (q - .5))

    def fit(self, x):
        x0 = np.median(stats.trimboth(x, .38))
        inner_quart = stats.trimboth(x, .25)
        inner_range = inner_quart.max() - inner_quart.min()
        return self.fit_ml(x, x0=(x0, .5 * inner_range))

    def _constraints(self, x, x0, gamma):
        if np.any(gamma < 0):
            return False
        return True
cauchy = Cauchy()
cauchy.parameter_names = ("x0", "gamma")
cauchy._feasible_start = (0, 1)


class Student_t(MyDist):
    def pdf(self, x, mu=0, df=1):
        x, mu, df = map(np.atleast_1d, (x, mu, df))
        Px = (np.exp(gammaln((df + 1) / 2) - gammaln(df / 2)) /
              np.sqrt(df * np.pi) *
              (1 + (x - mu) ** 2 / df) ** (-(df + 1) / 2))
        return np.asscalar(Px) if len(Px) == 1 else Px

    def cdf(self, x, mu=0, df=1):
        q = stdtr(df, np.atleast_1d(x) - mu)
        return np.asscalar(q) if len(q) == 1 else q

    def ppf(self, q, mu=0, df=1):
        x = stdtrit(df, np.atleast_1d(q)) + mu
        x[q == 0] = -np.inf
        x[q == 1] = np.inf
        return np.asscalar(x) if len(x) == 1 else x

    def fit(self, x):
        x_noninf = x[np.isfinite(x)]
        sigma = np.var(x_noninf)
        df = 2 * sigma / (sigma - 1)
        if df > 6:
            df = 6
        return self.fit_ml(x_noninf, x0=(np.mean(x_noninf),
                                         df if df > 0 else .1))

    def _constraints(self, x, mu, df):
        x, mu, df = (np.asarray(var) for var in (x, mu, df))
        if np.any((df <= 0) | (df > 8)):
            return False
        return True
student_t = Student_t()
student_t.parameter_names = ("mu", "df")
student_t._feasible_start = (0, 1)


class Noncentral_T(MyDist):
    # stolen from scipy.stats.distributions.nct
    def pdf(self, x, df, nc, mu=0):
        x, df, nc = (np.atleast_1d(var).astype(float) for var in (x, df, nc))
        pdf_x = lambda x: ((df / x) *
                           (self.cdf(x * (1 + 2 / df) ** .5, df + 2, nc) -
                            self.cdf(x, df, nc)))
        dens = np.where(np.abs(x - mu) <= .5,
                        self.pdf_old(x, df, nc, mu),
                        pdf_x(x - mu))
        dens[np.isinf(x)] = 0
        return np.asscalar(dens) if len(dens) == 1 else dens

    def pdf_old(self, x, df, nc, mu=0):
        n = df * 1.0
        nc = nc * 1.0
        x2 = (x - mu) ** 2
        ncx2 = nc * nc * x2
        fac1 = n + x2
        trm1 = (n / 2. * np.log(n) + gammaln(n + 1) -
                (n * np.log(2) + nc * nc / 2. + (n / 2.) * np.log(fac1) +
                 gammaln(n / 2.)))
        Px = np.exp(trm1)
        valF = ncx2 / (2 * fac1)
        trm1 = (np.sqrt(2) * nc * (x - mu) * hyp1f1(n / 2 + 1, 1.5, valF) /
                (np.asarray(fac1 * gamma_func((n + 1) / 2))))
        trm2 = (hyp1f1((n + 1) / 2, 0.5, valF) /
                (np.asarray(np.sqrt(fac1) * gamma_func(n / 2 + 1))))
        Px *= trm1 + trm2
        return Px

    def cdf(self, x, df, nc, mu=0):
        x, df, nc, mu = (np.atleast_1d(var).astype(float)
                         for var in (x, df, nc, mu))
        q = nctdtr(df, nc, x - mu)
        return np.asscalar(q) if len(q) == 1 else q

    def ppf(self, q, df, nc, mu=0):
        q, df, nc = (np.atleast_1d(var).astype(float) for var in (q, df, nc))
        x = nctdtrit(df, nc, q) + mu
        x[q == 0] = -np.inf
        x[q == 1] = np.inf
        return np.asscalar(x) if len(x) == 1 else x

    def fit(self, x):
        x_noninf = x[np.isfinite(x)]
        mode = stats.mode(np.round(x_noninf, 1))[0][0]
        x0 = (1, 2 * stats.skew(x_noninf), mode)
        df, nc, mu = self.fit_fsum(x_noninf, x0=x0)
#        if df >= 30:
#            # then we are normal anyway
#            df = 30
        return df, nc, mu

    def _constraints(self, x, df, nc, mu):
        df = np.asarray(df).astype(float)
        if np.any((df < .1) | (df > 8)):
            return False
        return True
noncentral_t = Noncentral_T()
noncentral_t.parameter_names = ("df", "nc", "mu")
noncentral_t._feasible_start = (2, 2, 0)


# TODO: implement ppf
# class LogitNormal(MyDist):
#    def logit(self, x):
#        return np.log(x / (1 - x))
#
#    def pdf(self, x, mu, sigma, l=0, u=1):
#        x_normed = (x - l) / (u - l)
#        logit = self.logit(x_normed)
#        return norm_pdf(logit, mu, sigma) / (x_normed * (1 - x_normed))
#
#    def cdf(self, x, mu, sigma, l=0, u=1):
#        x_normed = (x - l) / (u - l)
#        logit = self.logit(x_normed)
#        return norm_cdf(logit, mu, sigma)
#
#    def ppf(self, q, mu, sigma, l=0, u=1):
#        raise NotImplementedError
#
#    def fit(self, x, l=None, u=None):
#        l = x.min() - 1e-9 if l is None else l
#        u = x.max() + 1e-9 if u is None else u
#        x_normed = (x - l) / (u - l)
#        logit = self.logit(x_normed)
#        mu, sigma = norm_fit(logit)
#        return mu, sigma, l, u
#
#    def _constraints(self, x, mu, sigma, l=0, u=1):
#        x, mu, sigma, l, u = (np.asarray(var) for var in (x, mu, sigma, l, u))
#        if np.any(x < l):
#            return False
#        if np.any(x > u):
#            return False
#        return True
# logitnormal = LogitNormal()
# logitnormal.parameter_names = ("mu", "sigma", "l", "u")


# class Rayleigh(MyDist):
#    def pdf(self, x, sigma):
#        return (x / sigma ** 2) * np.exp(-x ** 2 / (2 * sigma ** 2))
#
#    def cdf(self, x, sigma):
#        return 1 - np.exp(-x ** 2 / (2 * sigma ** 2))
#
#    def fit(self, x):
#        return ((.5 * np.average(x ** 2)) ** .5,)
#
#    @abstractmethod
#    def ppf(self):
#        raise NotImplementedError
#
#    def _constraints(self, x, sigma):
#        if np.any(sigma <= 0):
#            return False
#        return True
# rayleigh = Rayleigh()
# rayleigh.parameter_names = ("sigma",)


# TODO: implement ppf
# class RayleighLU(MyDist):
#    def pdf(self, x, sigma, l, u):
#        x_normed = (x - l) / (u - l)
#        return (x_normed / sigma ** 2) * \
#            np.exp(-x_normed ** 2 / (2 * sigma ** 2))
#
#    def cdf(self, x, sigma, l, u):
#        x_normed = (x - l) / (u - l)
#        return 1 - np.exp(-x_normed ** 2 / (2 * sigma ** 2))
#
#    def fit(self, x, l=None, u=None):
#        l = x.min() if l is None else l
#        u = x.max() if u is None else u
#        x_normed = (x - l) / (u - l)
#        return (.5 * np.average(x_normed ** 2)) ** .5, l, u
#
#    def ppf(self):
#        raise NotImplementedError
#
#    def _constraints(self, x, sigma, l, u):
#        if np.any(sigma <= 0):
#            return False
#        if np.any(l >= u):
#            return False
#        return True
# rayleighlu = RayleighLU()
# rayleighlu.parameter_names = ("sigma", "l", "u")


# TODO: fit does not work well
# class LogNormalLU(MyDist):
#    def pdf(self, x, mu, sigma, l, u):
#        x_normed = (x - l) / (u - l)
#        return norm_pdf(np.log(x_normed + 1e-9), mu, sigma) / (u - l)
#
#    def cdf(self, x, mu, sigma, l, u):
#        x_normed = (x - l) / (u - l)
#        return norm_cdf(np.log(x_normed + 1e-9), mu, sigma)
#
#    def ppf(self, quantiles, mu, sigma, l, u):
#        return np.exp(norm_ppf(quantiles, mu, sigma)) * (u - l) + l
#
#    def fit(self, x, l=None, u=None):
#        x_noninf = x[np.isfinite(x)]
#        l = x_noninf.min() if l is None else l
#        u = x_noninf.max() if u is None else u
#        x_normed = (x_noninf - l) / (u - l)
#        mu, sigma = norm_fit(np.log(x_normed + 1e-9))
#        return self.fit_ml(x_noninf, x0=(mu, sigma, l, u))
#
#    def _constraints(self, x, mu, sigma, l, u):
#        if np.any(l >= u):
#            return False
#        return True
# lognormallu = LogNormalLU()
# lognormallu.parameter_names = ("mu", "sigma", "l", "u")
# lognormallu._feasible_start = (0, 1, -1, 1)


# class WeibullLU(MyDist):
#    def pdf(self, x, alpha, beta, l=0, u=1):
#        x_normed = (x - l) / (u - l)
#        # alpha and beta should be positive
#        densities = alpha * beta * x_normed ** (beta - 1) * \
#            np.exp(-alpha * x_normed ** beta)
#        return densities / (u - l)
#
#    def cdf(self, x, alpha, beta, l=0, u=1):
#        x_normed = (x - l) / (u - l)
#        return 1 - np.exp(-alpha * x_normed ** beta)
#
#    def ppf(self, quantiles, alpha, beta, l=0, u=1):
#        x = (-np.log(1 - quantiles) / alpha) ** (1 / beta)
#        return x * (u - l) + l
#
#    def fit(self, values, beta_start=10, *args, **kwds):
#        """Implements a semi-analytical method of moments. beta is found by
#        minimizing errors. After that, alpha can be estimated with the help of
#        beta.
#        See
#        http://interstat.statjournals.net/YEAR/2000/articles/0010001.norm_pdf
#        """
#        values_noninf = values[np.isfinite(values)]
#        l = values_noninf.min()
#        u = values_noninf.max()
#        values_noninf = (values_noninf - l) / (u - l)
#        # we need the coefficient of variation now and the mean later. so do
#        # not use stats.variation in order to not calculate the mean twice.
#        mean = np.mean(values_noninf)
#        cv = np.std(values_noninf) / mean
#
#        def cv_error(beta):
#            """Squared error between the theoretical and empirical coefficient
#            of variation."""
#            gamma1 = special.gamma(1 + 2 / beta)
#            gamma2 = special.gamma(1 + 1 / beta)
#            return (cv - (gamma1 - gamma2 ** 2) ** .5 / gamma2) ** 2
#        beta = sp_optimize.fmin(cv_error, [beta_start], disp=False)[0]
#        alpha = (mean / special.gamma(1 / beta + 1)) ** (-beta)
#        return alpha, beta, l, u
#
#    def _constraints(self, x, alpha, beta, l, u):
#        x, alpha, beta = (np.asarray(var) for var in (x, alpha, beta))
#        if np.any(alpha <= 0):
#            return False
#        if np.any(beta <= 0):
#            return False
#        if np.any(x < l):
#            return False
#        if np.any(x > u):
#            return False
#        return True
# weibulllu = WeibullLU()
# weibulllu.parameter_names = ("alpha", "beta", "l", "u")
# weibulllu._feasible_start = (1, 1, -1, 1)


class Weibull(MyDist):
    def pdf(self, x, alpha, beta):
        # alpha and beta should be positive
        densities = alpha * beta * x ** (beta - 1) * \
            np.exp(-alpha * x ** beta)
        return densities

    def cdf(self, x, alpha, beta):
        return 1 - np.exp(-alpha * x ** beta)

    def ppf(self, quantiles, alpha, beta):
        return (-np.log(1 - quantiles) / alpha) ** (1 / beta)

    def fit(self, values, beta_start=10, *args, **kwds):
        """Implements a semi-analytical method of moments. beta is found by
        minimizing errors. After that, alpha can be estimated with the help of
        beta.
        See
        http://interstat.statjournals.net/YEAR/2000/articles/0010001.norm_pdf
        """
        # we need the coefficient of variation now and the mean later. so do
        # not use stats.variation in order to not calculate the mean twice.
        mean = np.mean(values[np.isfinite(values)])
        cv = np.std(values[np.isfinite(values)]) / mean

        def cv_error(beta):
            """Squared error between the theoretical and empirical coefficient
            of variation."""
            gamma1 = special.gamma(1 + 2 / beta)
            gamma2 = special.gamma(1 + 1 / beta)
            return (cv - (gamma1 - gamma2 ** 2) ** .5 / gamma2) ** 2
        beta = sp_optimize.fmin(cv_error, [beta_start], disp=False)[0]
        alpha = (mean / special.gamma(1 / beta + 1)) ** (-beta)
        return alpha, beta

    def _constraints(self, x, alpha, beta):
        x, alpha, beta = (np.asarray(var) for var in (x, alpha, beta))
        if np.any(x < 0):
            return False
        if np.any(alpha <= 0):
            return False
        if np.any(beta <= 0):
            return False
        return True
weibull = Weibull()
weibull.parameter_names = ("alpha", "beta")
weibull._feasible_start = (1, 1)


class Kumaraswamy(MyDist):
    """Resembles the Beta distribution, but does not need a transcendental
    function."""
    def pdf(self, x, a, b, l=0, u=1):
        x = np.atleast_1d(x)
        if NE:
            x = ne.evaluate("(x - l) / (u - l)")
            dens = ne.evaluate("a * b * x ** (a - 1) * (1 - x ** a) ** " +
                               "(b - 1) / (u - l)")
        else:
            x = (x - l) / (u - l)
            dens = a * b * x ** (a - 1) * (1 - x ** a) ** (b - 1) / (u - l)
        return np.asscalar(dens) if len(dens) == 1 else dens

    def cdf(self, x, a, b, l=0, u=1):
        x = np.atleast_1d(x)
        if np.any((x < l) | (x > u)):
            warnings.warn("Some values below lower or above upper bounds.")
        q = 1 - (1 - ((x - l) / (u - l)) ** a) ** b
        return np.asscalar(q) if len(q) == 1 else q

    def ppf(self, quantiles, a, b, l=0, u=1):
        quantiles, a, b, l, u = map(np.atleast_1d, (quantiles, a, b, l, u))
        if NE:
            x = ne.evaluate("((1 - (1 - quantiles) ** (1 / b)) ** " +
                            "(1 / a)) * (u - l) + l")
        else:
            x = ((1 - (1 - quantiles) ** (1 / b)) ** (1 / a)) * (u - l) + l
        return np.asscalar(x) if len(x) == 1 else x

    def fit(self, x, l=None, u=None):
        return beta.fit(x, l, u)

    def _constraints(self, x, a, b, l=0, u=1):
        x, a, b, l, u = (np.asarray(var) for var in (x, a, b, l, u))
        if np.any(a < 0):
            return False
        if np.any(b < 0):
            return False
        if np.any(x < l):
            return False
        if np.any(x > u):
            return False
        return True
kumaraswamy = Kumaraswamy()
kumaraswamy.parameter_names = ("a", "b", "l", "u")
kumaraswamy._feasible_start = (1, 1, 0, 1)


class Beta(MyDist):
    """Beta distribution on the interval [l, u]."""
    def pdf(self, x, alpha, beta, l=0, u=1):
        x, alpha, beta, l, u = map(np.asarray, (x, alpha, beta, l, u))
        # putting the values into [0, 1] according to [l, u]
        # we avoid the value of exactly 0 or 1 to not get numerical problems
#        x_normed = (x - l) / (u - l)
#        xx = np.copy(x_normed)
#        ll, uu = sys.float_info.min, 1 - sys.float_info.min
#        x_normed = np.where(x_normed >= uu, uu, x_normed)
#        x_normed = np.where(x_normed <= ll, ll, x_normed)
        if NE:
            x = ne.evaluate("(x - l) / (u - l)")
            return 1 / special.beta(alpha, beta) * \
                ne.evaluate("x ** (alpha - 1) * (1 - x) ** (beta - 1) / " +
                            "(u - l)")
        else:
            x = (u - l) / (u - l)
            return (1 / special.beta(alpha, beta) * x ** (alpha - 1) * (1 - x)
                    ** (beta - 1) / (u - l))

    def cdf(self, x, alpha, beta, l=0, u=1):
        if np.any((x < l) | (x > u)):
            warnings.warn("Some values below lower or above upper bounds.")
        # note: the betainc function below really delivers the regularized
        # incomplete beta function
        return special.betainc(alpha, beta, (x - l) / (u - l))

    def ppf(self, q, alpha, beta, l=0, u=1):
        return (u - l) * special.betaincinv(alpha, beta, q) + l

    def fit(self, x, l=None, u=None):
        """stolen from
        http://en.wikipedia.org/wiki/Beta_distribution#Parameter_estimation."""
        # remember whether l and u where given. if they were, we do not return
        # them
        if l is None:
            l = x.min()
            return_l = True
        else:
            return_l = False
        if u is None:
            u = x.max()
            return_u = True
        else:
            return_u = False

        # have to make sure that the parameters all have the same lengths
        try:
            ll = np.empty_like(u)
            ll[:] = l
            l = ll
        except (TypeError, ValueError):
            pass
        try:
            uu = np.empty_like(l)
            uu[:] = u
            u = uu
        except (TypeError, ValueError):
            pass

        xmean = (x.mean() - l) / (u - l)
        xvar = x.var() / (u - l) ** 2
        alpha = xmean * (xmean * (1 - xmean) / xvar - 1)
        beta = (1 - xmean) * alpha / xmean

        return_list = [abs(alpha), abs(beta)]
        if return_l:
            return_list += [l]
        if return_u:
            return_list += [u]
        return return_list
#        return self.fit_ml(x, x0=(alpha, beta, l, u))

    def _constraints(self, x, alpha, beta, l=0, u=1):
        x, alpha, beta, l, u = (np.asarray(var)
                                for var in (x, alpha, beta, l, u))
        if np.any(alpha <= 0):
            return False
        if np.any(beta <= 0):
            return False
        if np.any(x < l):
            return False
        if np.any(x > u):
            return False
        return True
beta = Beta()
beta.parameter_names = ("alpha", "beta", "l", "u")
beta._feasible_start = (1, 1, 0, 1)


class Gamma(MyDist):
    """Gamma distribution."""
    def pdf(self, x, k, theta):
        x = np.atleast_1d(x)
        dens = (x ** (k - 1) * np.exp(-x / theta) /
                (special.gamma(k) * theta ** k))
        return np.asscalar(dens) if len(dens) == 1 else dens

    def cdf(self, x, k, theta):
        x = np.atleast_1d(x)
        q = gammainc(k, x / theta) / gamma_func(k)
        q[np.isinf(x)] = 1
        return np.asscalar(q) if len(q) == 1 else q

    def ppf(self, q, k, theta):
        q = np.atleast_1d(q)
        x = theta * gammaincinv(k, q * gamma_func(k))
        x[q == 1] = np.inf
        return np.asscalar(x) if len(x) == 1 else x

    def fit(self, x, rel_change=1e-6):
        """A maximum-likelihood estimator. See
        http://en.wikipedia.org/wiki/Gamma_distribution#Parameter_estimation
        """

        def psi(k):
            if k >= 8:
                return np.log(k) - (1 + (1 - (.1 - 1 / (21 * k ** 2)) /
                                         k ** 2) / (6 * k)) / (2 * k)
            else:
                return psi(k + 1) - 1. / k

        def psi_(k):
            if k >= 8:
                return (1 + (1 + (1 - (1. / 5 - 1 / (7 * k ** 2)) / k ** 2) /
                        (3 * k)) / (2 * k)) / k
            else:
                return psi_(k + 1) + 1 / k ** 2

        xmean = x[np.isfinite(x)].mean()
        # we get problems here if there are 0's in x (log(x) = -inf, which
        # happens for wind)
        s = np.log(xmean) - np.mean(np.log(x[np.isfinite(x) & (x > 0)]))
        k_old = (3 - s + ((s - 3) ** 2 + 24 * s) ** .5) / (12 * s)
        k_rec = lambda k: k - (np.log(k) - psi(k) - s) / (1. / k - psi_(k))
        k_new = k_rec(k_old)
        while abs(k_old - k_new) / k_old > rel_change:
            k_old = k_new
            k_new = k_rec(k_old)
        theta = xmean / k_new
        return self.fit_fsum(x, x0=(k_new, theta))

    def _constraints(self, x, k, theta):
        x, k, theta = (np.asarray(var) for var in (x, k, theta))
        if np.any(k <= 0):
            return False
        if np.any(theta <= 0):
            return False
        if np.any(x < 0):
            return False
        return True
gamma = Gamma()
gamma.parameter_names = ("k", "theta")
gamma._feasible_start = (2, 2)


class Gamma1(MyDist):
    def pdf(self, x, a, loc, scale):
        x, a, loc, scale = map(np.atleast_1d, (x, a, loc, scale))
        z = (x - loc) / scale
        dens = (np.exp((a - 1) * np.log(z) - z - gammaln(a))) / scale

        return np.asscalar(dens) if len(dens) == 1 else dens

    def cdf(self, x, a, loc, scale):
        x, a, loc, scale = map(np.atleast_1d, (x, a, loc, scale))
        z = (x - loc) / scale
        q = gammainc(a, z)
        q[np.isinf(x)] = 1
        return np.asscalar(q) if len(q) == 1 else q

    def ppf(self, q, a, loc, scale):
        q, a, loc, scale = map(np.atleast_1d, (q, a, loc, scale))
        x = gammaincinv(a, q) * scale + loc
        x[q == 1] = np.inf
        return np.asscalar(x) if len(x) == 1 else x

    def fit(self, x):
        return stats.gamma.fit(x[np.isfinite(x)])
        # this is stolen from sp.stats.gamma-gen and not understood
#        x_noninf = x[np.isfinite(x)]
#        a = 4 / stats.skew(x_noninf) ** 2
#        muhat = np.array(x_noninf).mean()
#        mu2hat = np.array(x_noninf).var()
#        Shat = (mu2hat / a) ** .5
#        Lhat = muhat - Shat * a
#        return self.fit_ml(x_noninf, x0=(a, Lhat, Shat))
gamma1 = Gamma1()
gamma1.parameter_names = ("a", "loc", "scale")
gamma1._feasible_start = (1, 1, 2)


class Expon(MyDist):
    def pdf(self, x, x0, lambd):
        x = np.atleast_1d(x)
        dens = lambd * np.exp(-lambd * (x - x0))
        dens[dens > 1] = 0
        return np.asscalar(dens) if len(dens) == 1 else dens

    def cdf(self, x, x0, lambd):
        q = 1 - np.exp(-lambd * (np.atleast_1d(x) - x0))
        return np.asscalar(q) if len(q) == 1 else q

    def ppf(self, q, x0, lambd):
        x = -np.log(1 - np.atleast_1d(q)) / lambd + x0
        return np.asscalar(x) if len(x) == 1 else x

    def fit(self, x):
        x_noninf = x[np.isfinite(x)]
        x0 = np.min(x_noninf)
        lambd = 1 / np.mean(x_noninf)
        return x0, lambd

    def _constraints(self, x, x0, lambd):
        if np.any(x <= x0):
            return False
        return True
expon = Expon()
expon.parameter_names = ("x0", "lambd")
expon._feasible_start = (0, 1)


class Expon_two(MyDist):
    def pdf(self, x, x0, q0, lambda1, lambda2):
        x = np.atleast_1d(x)
        dens = np.where(x <= x0,
                        q0 * expon.pdf(x0 - x, 0, lambda1),
                        (1 - q0) * expon.pdf(x, x0, lambda2))
        return np.asscalar(dens) if len(dens) == 1 else dens

    def cdf(self, x, x0, q0, lambda1, lambda2):
        x = np.atleast_1d(x)
        q = np.where(x <= x0,
                     q0 * (1 - expon.cdf(x0 - x, 0, lambda1)),
                     q0 + (1 - q0) * expon.cdf(x, x0, lambda2))
        return np.asscalar(q) if len(q) == 1 else q

    def ppf(self, q, x0, q0, lambda1, lambda2):
        q = np.atleast_1d(q)
        x = np.where(q <= q0,
                     x0 + np.log(q / q0) / lambda1,
                     x0 - np.log(1 - (q - q0) / (1 - q0)) / lambda2)
        return np.asscalar(x) if len(x) == 1 else x

    def fit(self, x):
        x_noninf = x[np.isfinite(x)]
        x0 = stats.mode(np.round(x_noninf, 1))[0][0]
        lambda1 = -1 / np.mean(x_noninf[x_noninf < x0] - x0)
        lambda2 = 1 / np.mean(x_noninf[x_noninf > x0] - x0)
        q0 = (np.argmin(np.abs(np.sort(x_noninf) - x0)) + .5) / len(x)
        return self.fit_ml(x_noninf, x0=(x0, q0, lambda1, lambda2))

    def _constraints(self, x, x0, q0, lambda1, lambda2):
        if np.any(lambda1 <= 0):
            return False
        if np.any(lambda2 <= 0):
            return False
        if np.any((q0 < 0) | (q0 > 1)):
            return False
        return True
expon_two = Expon_two()
expon_two.parameter_names = ("x0", "q0", "lambda1", "lambda2")
expon_two._feasible_start = (0, .5, 1, 1)


class Noncentral_Laplace(MyDist):
    def pdf(self, x, x0, q0, lambd):
        lambda2 = lambd * q0 / (1 - q0)
        x = np.atleast_1d(x)
        dens = np.where(x <= x0,
                        q0 * lambd * np.exp(-lambd * (x0 - x)),
                        (1 - q0) * lambda2 * np.exp(-lambda2 * (x - x0)),
                        )
        dens[np.isinf(x)] = 0
        return np.asscalar(dens) if len(dens) == 1 else dens

    def cdf(self, x, x0, q0, lambd):
        lambda2 = lambd * q0 / (1 - q0)
        x = np.atleast_1d(x)
        q = np.where(x <= x0,
                     q0 * (1 - expon.cdf(x0 - x, 0, lambd)),
                     q0 + (1 - q0) * expon.cdf(x, x0, lambda2))
        return np.asscalar(q) if len(q) == 1 else q

    def ppf(self, q, x0, q0, lambd):
        lambda2 = lambd * q0 / (1 - q0)
        q = np.atleast_1d(q)
        x = np.where(q <= q0,
                     x0 + np.log(q / q0) / lambd,
                     x0 - np.log(1 - (q - q0) / (1 - q0)) / lambda2)
        return np.asscalar(x) if len(x) == 1 else x

    def fit(self, x):
        x0 = stats.mode(np.round(x, 1))[0][0]
        q0 = (np.argmin(np.abs(np.sort(x) - x0)) + .5) / len(x)
        lambd = (1 - 2 * q0) / (q0 * np.mean(x[np.isfinite(x)]))
        if lambd < 0:
            # this actually means that x0 was not right, so provide a feasible
            # lambd and let the ml estimation figure the rest out
            lambd = -1. / np.mean(x[(x < x0) & np.isfinite(x)] - x0)
        return self.fit_ks(x, x0=(x0, q0, lambd))

    def _constraints(self, x, x0, q0, lambd):
        if np.any(lambd <= 0):
            return False
        if np.any((q0 < 0) | (q0 > 1)):
            return False
        return True
noncentral_laplace = Noncentral_Laplace()
noncentral_laplace.parameter_names = ("x0", "q0", "lambd")
noncentral_laplace._feasible_start = (1, .5, .5)


# def expon_pdf(x, loc, scale):
#    x, loc, scale = np.asarray(x), np.asarray(loc), np.asarray(scale)

if __name__ == "__main__":
    import vg
    dtimes_, met = vg.read_met()
    u, v = met["u"], met["v"]
    (u, dtimes), v = np.squeeze(vg.my.sumup(u, 24, dtimes_)), np.squeeze(vg.my.sumup(v, 24))
    u, v = u.astype(float) / 24, v.astype(float) / 24
    for month in range(1, 13):
        u1 = u[vg.times.time_part(dtimes, "%m") == month]
        u_dist = noncentral_laplace(*noncentral_laplace.fit(u1))
        u_dist.plot_qq(u1)
        plt.title("month: %d" % month)
        u_dist.plot_fit(u1)
        plt.title("month: %d" % month)
    plt.show()



#    R, dtimes = np.squeeze(vg.my.sumup(met["R"], 24, dtimes_))
#    R = R.astype(float)
#    R_dist = stats.gamma(*stats.gamma.fit(R))
#    R_dist.pdf(0)
#    cen_exp = vg.distributions.Censored(vg.distributions.expon, lc=0)
#    r_cen = cen_exp(*cen_exp.fit(R))
