""" the following homogeneity tests are described in a minimal manner in
Peterson et al, "Homogeneity Adjustments of In Situ Atmospheric Climate Data:
A Review

rocpt cannot detect changes in slope of trend (snht can), but in finding the
change from 'no trend' to 'trend', rocpt is better than snht
"""

import numpy as np
from scipy import stats

#------------------------------------------------------------------------------#
def cusum(values, k=np.mean, axis=None):
    return np.cumsum(values - k(values), axis=axis)
#------------------------------------------------------------------------------#
def snht(values):
    """Standard normal homogeneity test in the simplest form. Returns the index
    of the most possible breakpoint in values.

    References
    ----------
    Alexandersson, 'A homogeneity test applied to precipitation data'.
    Tested on the ERA-40 dataset, which has a breakpoint towards the end of
    1979."""
    n = len(values)
    v = np.arange(n)
    z1 = np.array([values[:ii].mean() for ii in v])
    z2 = np.array([values[ii:].mean() for ii in v])
    Tv = v * z1 ** 2 + (n - v) * z2 ** 2
    # throw away first and last value, because they could be nans and are
    # generally not useful in break point analysis anyway
    ii = np.argmax(Tv[1:-1]) + 1
    downwards = False
    if (z1 ** 2)[ii] > (z2 ** 2)[ii]: downwards = True
    return ii, downwards
#------------------------------------------------------------------------------#
def rocpt(values, sig=0.02):
    """Rank order change point test. Returns the index of the most possible
    breakpoint in values.

    Ranks -> cannot find changes in slope of trend. Use snht instead.

    Parameters
    ----------
    values : np.array
        time series of data
    sig : float, optional
        significance level

    Returns
    -------
    x : int
        index of the most possible breakpoint
    test_z : float
        test statistic from rocpt_test
    test_ : boole
        test result from rocpt_test

    References
    ----------
    Lanzante, 'Resistant, robust and nonparametric techniques for the
    analysis of climate data'.
    Tested on the ERA-40 dataset, which has a breakpoint towards the end of
    1979."""
    n = len(values)
    i = np.arange(n)+1 # ranks start with 1
    SRi = np.cumsum(stats.rankdata(values))
    SAi = np.abs(2 * SRi - i * (n + 1))
    x = np.argmax(SAi[:-1])
    SRx = SRi[x]
    test_z, test_ = rocpt_test(SRx, x, n, sig=sig)
    return x, test_z, test_
#------------------------------------------------------------------------------#
def rocpt_test(SRx, x, n, sig=0.02):
    """Test statistic for rank order change point test.

    References
    ----------
    Peterson et al."""
    n, x = int(n), int(x)
    d = 0.5 # if SRx < x*(n+1)/2
    if SRx > x*(n+1)/2.: d = -0.5
    elif SRx == x*(n+1)/2.: d = 0.
    oben = SRx - x*(n+1)/2. + d
    unten = ( x * (n-x) * (n+1) / 12.)**0.5
    z = oben/unten
    test_=False
    if z>stats.norm.ppf(1.-sig/2.): test_=True
    if z<stats.norm.ppf(sig/2.): test_=True
    if ~np.isfinite(z): test_=False
    return z, test_
#------------------------------------------------------------------------------#