"""Tools to fit seasonal distributions by describing the yearly cycle of
distribution parameters with triangular functions."""
import numpy as np
import scipy as sp
import matplotlib.pyplot as plt
from vg.time_series_analysis import optimize
from vg.time_series_analysis import distributions
from vg import times
from vg import smoothing
import vg.helpers as my


class SeasonalDist(distributions.MyDist):
    def __init__(self, distribution, x, datetimes, fixed_pars=None,
                 par_ntrig=None, time_form="%m", verbose=False):
        """distribution should be an object that implements pdf, cdf and so on.
        fixed_pars is expected to be a dictionary mapping parameter names to
        functions that take days of the year as input to calculate distribution
        parameters.
        time_form (e.g. "%m") determines what is used to generate a starting
        solution of trigonometric parameters.
        var_ntrig should be a sequence mapping distribution parameters to
        number of trigonometric parameters. If that is None, 3 trig  parameters
        per dist parameter are assumed
        """
        self.verbose = verbose
        # check if this is a generalized truncated distribution
        try:
            self.dist = distribution[0](distribution[1])
        except TypeError:
            self.dist = distribution
        finite_ii = np.isfinite(x)
        self.x = x[finite_ii]
        self.times_ = datetimes[finite_ii]
        self.doys = times.datetime2doy(self.times_)
        self.fixed_pars = {} if fixed_pars is None else fixed_pars
        if not hasattr(self.dist, "parameter_names"):
            # we assume that we have come across a scipy.stats.distribution
            self.dist.parameter_names = ("loc", "scale")
        if par_ntrig is None:
            self.par_ntrig = [3] * len(self.dist.parameter_names)
        else:
            self.par_ntrig = par_ntrig
        self.time_form = time_form
        self.non_fixed_names = [par_name for par_name
                                in self.dist.parameter_names
                                if par_name not in self.fixed_pars.keys()]
        self.n_trig_pars = 3
        self.pre_start = []

    def trig2pars(self, trig_pars, _T=None):
        """This is the standard "give me distribution parameters for
        trigonometric parameters" that assumes 3 trig pars per distribution
        parameter."""
        if _T is None:
            try:
                _T = self._T
            except AttributeError:
                _T = self._T = (2 * np.pi / 365 * self.doys)[np.newaxis, :]
        # funny how the obvious thing looks so complicated
        a, b_0, phi = \
            np.array(trig_pars).reshape(-1, 3).T[:, :, np.newaxis]
        # i lied, this took me some time. but now, the parameters are vertical
        # to T_ and we are broadcasting like some crazy shit pirate radio
        # station
        return a + b_0 * np.sin(_T + phi)

    def fixed_values_dict(self, doys=None):
        """This holds a cache for the self.doys values. If doys is given, those
        are used and the cache is left untouched."""
        def build_dict(doys):
            return dict((fixed_par_name, func(doys))
                        for fixed_par_name, func in self.fixed_pars.items())
        if doys is None:
            try:
                return self._cached_fixed
            except AttributeError:
                self._cached_fixed = build_dict(self.doys)
            return self.fixed_values_dict()
        else:
            return build_dict(doys)

    def all_parameters_dict(self, trig_pars, doys=None):
        """Returns a dictionary of both the fixed and fitted (or to be fitted)
        distribution paramters."""
        # copy (!) the fixed_values into params
        params = dict((key, val) for key, val in
                      self.fixed_values_dict(doys).items())
        if doys is not None:
            doys = np.copy(doys) * 2 * np.pi / 365
        params.update(
            dict((param_name, values) for param_name, values
                 in zip(self.non_fixed_names,
                        self.trig2pars(trig_pars, doys))))
        return params

    @property
    def monthly_grouped(self):
        try:
            return self._monthly_grouped
        except AttributeError:
            self._monthly_grouped = times.time_part_sort(self.times_, self.x,
                                                         self.time_form)
        return self.monthly_grouped

    @property
    def monthly_grouped_x(self):
        try:
            return self._monthly_grouped[1]
        except AttributeError:
            # calling monthly grouped has the side effect of also calculating
            # monthly_grouped_x
            self.monthly_grouped
        return self.monthly_grouped[1]

    @property
    def daily_grouped_x(self):
        try:
            return self._daily_grouped_x
        except AttributeError:
            self._daily_grouped_x = \
                times.time_part_sort(self.times_, self.x, "%j")[1]
        return self.daily_grouped_x

    @property
    def monthly_params(self):
        """These are the distribution parameters per month. Returns a
        12xn_distribution_parameters array."""
        return np.array([self.dist.fit(
            sub_x,
            **dict((par_name,
                    np.average(func(self.doys[times.time_part(self.times_,
                                                              self.time_form)
                                              == month])))
                   for par_name, func in self.fixed_pars.items()))
            for month, sub_x in zip(*self.monthly_grouped)])

    def get_start_params(self, opt_func=sp.optimize.fmin, x0=None, **kwds):
        """Estimate seasonal distribution parameters by fitting a
        trigonometric function to monthly lumped data. Lower and upper bounds
        (if they are defined in the distribution) will be fitted conservatively
        to at least ensure that the start parameters are a feasible solution.
        """
        if x0 is None:
            x0 = [0, 1, 0] + self.pre_start
        if "maxfun" not in kwds:
            kwds["maxfun"] = 1e5 * sum(self.par_ntrig)
        #TODO: friggin if shit
        if self.time_form == "%W":
            _T = np.linspace(0, 365, 54, endpoint=False)
        elif self.time_form == "%m":
            first_doms = times.str2datetime("01.%02d.2000 00:00:00" % month
                                            for month in range(1, 13))
            _T = np.array(times.time_part(first_doms, "%j"), dtype=float)
        _T *= 2 * np.pi / 365

        def error(trig_pars, emp_pars, times_=None):
            times_ = _T if times_ is None else times_
            return np.sum((self.trig2pars(trig_pars, times_) - emp_pars) ** 2)

        start_pars = []
        for par_name, par_timeseries in zip(self.dist.parameter_names,
                                            self.monthly_params.T):
            # if the distribution comes with lower and upper bounds, find a
            # solution for those that contains all the measurement points.
            if par_name == "l" and "l" not in self.fixed_pars:
                minima = np.array([var.min()
                                   for var in self.daily_grouped_x])
                smoothed_mins = smoothing.smooth(minima, 365, periodic=True)
                # place the phase shift so that the minimum of the sine is at
                # the the minimum of the dataset
                mint = self.doys[np.argmin(smoothed_mins)]
                phi0 = 1.5 * np.pi - mint * 2 * np.pi / 365
                b_00 = .5 * (smoothed_mins.max() - smoothed_mins.min())
                a0 = smoothed_mins.mean()
                start = [a0, b_00, phi0] + self.pre_start
                lowest_diff = np.min(smoothed_mins - minima)
                k = 1.5
                while np.any(smoothed_mins < k * lowest_diff):
                    k += .5
                fit_me = smoothed_mins - k * lowest_diff
                times_ = np.arange(366.) * 2 * np.pi / 365
                start_pars += list(opt_func(error, start,
                                            args=(fit_me, times_), disp=False,
                                            **kwds))
            elif par_name == "u" and "u" not in self.fixed_pars:
                maxima = np.array([var.max()
                                   for var in self.daily_grouped_x])
                smoothed_maxs = smoothing.smooth(maxima, 365, periodic=True)
                maxt = self.doys[np.argmax(smoothed_maxs)]
                phi0 = .5 * np.pi - maxt * 2 * np.pi / 365
                b_00 = .5 * (smoothed_maxs.max() - smoothed_maxs.min())
                a0 = smoothed_maxs.mean()
                start = [a0, b_00, phi0] + self.pre_start
                highest_diff = np.max(maxima - smoothed_maxs)
                fit_me = smoothed_maxs + 1.2 * highest_diff
                times_ = np.arange(366.) * 2 * np.pi / 365
                start_pars += list(opt_func(error, start, disp=False,
                                            args=(fit_me, times_), **kwds))
            elif par_name not in self.fixed_pars:
                start_pars += list(opt_func(error, x0, args=(par_timeseries,),
                                            disp=False, **kwds))
        self._start_pars = start_pars
        return start_pars

    def _complete_dist_call(self, func, trig_pars, x=None, doys=None,
                            broadcast=False, **kwds):
        """Abstracts the common ground for self.{pdf,cdf,ppf} which call the
        according functions of the underlying distribution."""
        params = self.all_parameters_dict(trig_pars, doys)
        params.update(kwds)
        x = self.x if x is None else x
        if broadcast:
            xx = x[np.newaxis, :]
            params = dict((par_name, val[:, np.newaxis])
                          for par_name, val in params.items())
            return func(xx, **params)
        else:
            return func(x, **params)

    def pdf(self, trig_pars, x=None, doys=None, **kwds):
        return self._complete_dist_call(self.dist.pdf, trig_pars, x, doys,
                                        **kwds)

    def cdf(self, trig_pars, x=None, doys=None, **kwds):
        return self._complete_dist_call(self.dist.cdf, trig_pars, x, doys,
                                        **kwds)

    def ppf(self, trig_pars, quantiles=None, doys=None, **kwds):
        return self._complete_dist_call(self.dist.ppf, trig_pars, quantiles,
                                        doys, **kwds)

    @property
    def solution(self):
        try:
            return self._solution
        except AttributeError:
            # self.fit sets the self._solution attribute
            self.fit()
            return self.solution

    @solution.setter
    def solution(self, sol):
        self._solution = sol

    @property
    def start_pars(self):
        try:
            return self._start_pars
        except AttributeError:
            self._start_pars = self.get_start_params()
        return self.start_pars

    def fit(self, x=None, opt_func=optimize.simulated_annealing, x0=None,
            **kwds):
        if x is not None:
            self.x = x
        if x0 is None:
            x0 = self.get_start_params(maxfun=1e3 * sum(self.par_ntrig))

        def constraints(trig_pars):
            dist_params = self.all_parameters_dict(trig_pars)
            return self.dist._constraints(self.x, **dist_params)

        def unlikelihood(trig_pars):
            densities = self.pdf(trig_pars) + 1e-12
            obj_value = -np.sum(np.log(densities))
            if not np.isfinite(obj_value):
                raise ValueError("Non-finite objective function value.")
            return obj_value

        def chi2(trig_pars):
            quantiles = self.cdf(trig_pars)
            f_obs = np.histogram(quantiles, 40)[0].astype(float)
            f_obs /= f_obs.sum()
            f_exp = np.array([float(len(f_obs))] * len(f_obs))
            obj_value = sp.stats.chisquare(f_obs, f_exp)[0]
            if not np.isfinite(obj_value):
                raise (ValueError, "Non-finite objective function value")
            return obj_value

#        n = len(self.x)
#        x_sorted_ii = np.argsort(self.x)
#        ranks_plus = np.arange(0., n) / n
#        ranks_minus = np.arange(1., n + 1) / n
#        def ks(trig_pars):
#            cdf_values = self.cdf(trig_pars)[x_sorted_ii]
#            #cdf_values[np.isnan(cdf_values)] = np.inf
#            dmin_plus = np.abs(cdf_values - ranks_plus).max()
#            dmin_minus = np.abs(cdf_values - ranks_minus).max()
#            return max(dmin_plus, dmin_minus)
#        def f_diff(trig_pars):
#            cdf_values = self.cdf(trig_pars)[x_sorted_ii]
#            #cdf_values[np.isnan(cdf_values)] = np.inf
#            dmin_plus = np.sum((cdf_values - ranks_plus) ** 2)
#            dmin_minus = np.sum((cdf_values - ranks_minus) ** 2)
#            return dmin_plus + dmin_minus
#        def combined(trig_pars):
#            return unlikelihood(trig_pars) + f_diff(trig_pars)
        result = opt_func(unlikelihood, x0, constraints=(constraints,),
                          callback=lambda x: setattr(self, "_solution", x),
                          **kwds)
#        try:
#            result = opt_func(unlikelihood, x0, constraints=(constraints,),
#                              callback=lambda x: setattr(self, "_solution",
#                                                         x),
#                              **kwds)
#        except ValueError:
#            #HACK!
#            result = x0
        if opt_func is sp.optimize.anneal:
            print "retval was %d" % result[-1]
            result = result[0]
        self._solution = result
        return result

    def scatter_cdf(self, trig_pars=None, n_sumup=24, figsize=None,
                     title=None, *args, **kwds):
        if trig_pars is None:
            try:
                trig_pars = self._solution
            except AttributeError:
                trig_pars = self.start_pars
        doys = np.arange(1, 366, dtype=float)
        _T = doys * 2 * np.pi / 365
        xx = np.linspace(self.x.min(), self.x.max(), 100)
        quants = self.cdf(trig_pars, xx, doys, broadcast=True)
        fig = plt.figure(figsize=figsize)
        ax1 = fig.gca()
        co = ax1.contourf(doys, xx / n_sumup, quants.T, 15)
        plt.scatter(self.doys, self.x / n_sumup, marker="o",
                   facecolors=(0, 0, 0, 0), edgecolors=(0, 0, 0, .5))
        # plot lower and upper bounds
        for fixed_values in self.fixed_values_dict(doys).values():
            plt.plot(doys, fixed_values / n_sumup, "r")
        try:
            for par in (trig_pars[2 * self.n_trig_pars:
                                  3 * self.n_trig_pars],
                        trig_pars[3 * self.n_trig_pars:
                                  4 * self.n_trig_pars]):
                plt.plot(doys, np.squeeze(self.trig2pars(par, _T)), "r")
        except (IndexError, ValueError):
            pass
        plt.colorbar(co)
        plt.xlim(0, len(doys))
        plt.ylim(xx.min() / n_sumup, xx.max() / n_sumup)
        plt.xlabel("Day of Year")
        plt.grid()
        plt.legend()
        if title is not None:
            plt.title(title)
        return fig

    def scatter_pdf(self, trig_pars=None, n_sumup=24, figsize=None,
                    title=None, opacity=.25, *args, **kwds):
        if trig_pars is None:
            try:
                trig_pars = self._solution
            except AttributeError:
                trig_pars = self.start_pars
        doys = np.unique(self.doys)
        _T = doys * 2 * np.pi / 365
        xx = np.linspace(self.x.min(), self.x.max(), 100)
        dens = self.pdf(trig_pars, xx, doys, broadcast=True)
        #dens[dens < 1e-12] = 0
        fig = plt.figure(figsize=figsize)
        ax1 = fig.gca()
        co = ax1.contourf(doys, xx / n_sumup, dens.T, 15)
#        plt.colorbar(co)
        plt.scatter(self.doys, self.x / n_sumup, marker="o",
                    facecolors=(0, 0, 0, 0), edgecolors=(0, 0, 0, opacity))
#        if self.dist in (distributions.beta, distributions.kumaraswamy,
#                         distributions.Kumaraswamy,
#                         distributions.weibull, distributions.logitnormal,
#                         distributions.lognormallu, distributions.rayleighlu):
        # plot lower and upper bounds
        for fixed_values in self.fixed_values_dict(doys).values():
            plt.plot(doys, fixed_values / n_sumup, "r")
        try:
            for par in (trig_pars[2 * self.n_trig_pars:
                                  3 * self.n_trig_pars],
                        trig_pars[3 * self.n_trig_pars:
                                  4 * self.n_trig_pars]):
                plt.plot(doys, np.squeeze(self.trig2pars(par, _T)) / n_sumup,
                         "r")
        except (IndexError, ValueError):
            pass

#        if self.dist is distributions.norm:
#            # do not plot the mean (boring!)
#            ax = plt.gca().twinx()
#            par = self.trig2pars(trig_pars[self.n_trig_pars:
#                                           2 * self.n_trig_pars], _T)
#            par = par.reshape(doys.shape)
#            plt.plot(doys, par, "r", label=self.dist.parameter_names[1])
#        else:
        # plot the first two distribution parameters on a different yscale

#        ax = plt.gca().twinx()
#        for par_i, par in enumerate((trig_pars[:self.n_trig_pars],
#                                     trig_pars[self.n_trig_pars:
#                                               2 * self.n_trig_pars])):
#            ax.plot(doys,
#                    self.trig2pars(par, _T).reshape(doys.shape) / n_sumup,
#                    label=self.dist.parameter_names[par_i])
        plt.xlim(0, 366)
        plt.ylim(xx.min() / n_sumup, xx.max() / n_sumup)
        plt.xticks((1, 32, 60, 91, 121, 152, 182, 213, 244, 274, 305, 335),
                   ('Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug',
                    'Sep', 'Oct', 'Nov', 'Dec'), rotation=45)
#        plt.xlabel("Day of Year")
        plt.grid()
#        plt.legend()
        if title is not None:
            plt.title(title)
        #plt.tight_layout()
        return fig

    def plot_seasonality_fit(self, solution=None):
        if self.time_form == "%m":
            doys = np.array(
                times.time_part(times.str2datetime(("2011 %d" % month
                                              for month in range(1, 13)),
                                              "%Y %m"),
                            "%j"), dtype=float)
        elif self.time_form == "%W":
            doys = \
                times.datetime2doy(
                    times.str2datetime(["2011 %d" % week
                                        for week in range(0, 54)],
                                       "%Y %W"))

        _T = 2 * np.pi * doys / 365
        if solution is None:
            solution = self.start_pars
        dist_params = self.trig2pars(solution, _T)
        for par_name, emp_pars, fit_pars in zip(self.dist.parameter_names,
                                                self.monthly_params.T,
                                                dist_params):
            plt.figure()
            plt.plot(emp_pars, label="empirical")
            plt.plot(fit_pars, label="fitted")
            plt.title(par_name)
            plt.legend()
        plt.show()

    def plot_monthly_fit(self, solution=None, n_classes=30):
        if solution is None:
            monthly_params = self.monthly_params
        else:
            # get the doys of the middle of the months very roughly
            month_doys = np.linspace(15, 367, 12,
                                     endpoint=False).astype(np.int)
            # get a parameter set per month
            monthly_params_dict = \
                self.all_parameters_dict(solution, month_doys)
            monthly_params = \
                np.array([monthly_params_dict[par_name]
                          for par_name in self.dist.parameter_names]).T
        for ii, values in enumerate(self.monthly_grouped_x):
            fig = plt.figure()
            ax1 = fig.add_subplot(111)

            # the histogram of the data
            bins = ax1.hist(values, n_classes, normed=True, facecolor='grey',
                            alpha=0.75)[1]

            class_middles = 0.5 * (bins[1:] + bins[:-1])
            monthly_dict = dict((par_name, pars) for par_name, pars in
                                zip(self.dist.parameter_names,
                                    monthly_params[ii]))
            density = self.dist.pdf(class_middles, **monthly_dict)
            # scale density, so both plots can be seen
            ax1.plot(class_middles, density, 'r--')

            # the quantile part
            ax2 = ax1.twinx()
            # empirical cdf
            values_sort = np.sort(values)
            ranks_emp = (.5 + np.arange(len(values))) / len(values)
            ax2.plot(values_sort, ranks_emp)
            # theoretical cdf
            xx = np.linspace(values.min(), values.max(), 5e2)
            ranks_theory = self.dist.cdf(xx, **monthly_dict)
            ax2.plot(xx, ranks_theory, 'r--')
            plt.title("month:%d %s" %
                     (ii + 1, ", ".join(" % s: %.3f" % (par_name, par)
                                    for par_name, par in
                                    zip(self.dist.parameter_names,
                                        monthly_params[ii]))))
            fig.canvas.set_window_title("month: %d (%d)" %
                                        (ii + 1, fig.canvas.manager.num))

    def plot_monthly_params(self):
        for par_name, values in zip(self.dist.parameter_names,
                                    self.monthly_params.T):
            plt.figure()
            plt.plot(values)
            plt.title(par_name)
        plt.show()


class SeasonalDist6d(SeasonalDist):
    def __init__(self, *args, **kwds):
        super(SeasonalDist6d, self).__init__(*args, **kwds)
        self.n_trig_pars = 6
        self.pre_start = [1, 1, 0]

    def trig2pars(self, trig_pars, _T=None):
        if _T is None:
            try:
                _T = self._T
            except AttributeError:
                _T = self._T = (2 * np.pi / 365 * self.doys)[np.newaxis, :]
        try:
            # funny how the obvious thing looks so complicated
            a, b_0, phi, b2, per, phi2 = \
                np.array(trig_pars).reshape(-1, 6).T[:, :, np.newaxis]
            # allow only integer multiples of 1/year frequency
            _T2 = _T * 365 / per.round().astype(int)
            # i lied, this took me some time. but now, the parameters are
            # vertical to T_ and we are broadcasting like some crazy shit
            # pirate radio station
            return a + b_0 * np.sin(_T + phi) + b2 * np.sin(_T2 + phi2)
        except ValueError:
            # try with 3 parameters (happens during get_start_params
            a, b_0, phi = \
                np.array(trig_pars).reshape(-1, 3).T[:, :, np.newaxis]
            return a + b_0 * np.sin(_T + phi)


class SeasonalDist4d(SeasonalDist):
    def __init__(self, *args, **kwds):
        super(SeasonalDist4d, self).__init__(*args, **kwds)
        self.n_trig_pars = 4
        self.pre_start = [1]

    def trig2pars(self, trig_pars, _T=None):
        if _T is None:
            try:
                _T = self._T
            except AttributeError:
                _T = self._T = (2 * np.pi / 365 * self.doys)[np.newaxis, :]
        try:
            # funny how the obvious thing looks so complicated
            a, b_0, phi, a2 = \
                np.array(trig_pars).reshape(-1, 4).T[:, :, np.newaxis]
            # i lied, this took me some time. but now, the parameters are
            # vertical to T_ and we are broadcasting like some crazy shit
            # pirate radio station
            return a - b_0 * np.cos(_T + phi) * (a2 - np.cos(_T + phi))
        except ValueError:
            # try with 3 parameters (happens during get_start_params
            a, b_0, phi = \
                np.array(trig_pars).reshape(-1, 3).T[:, :, np.newaxis]
            return a + b_0 * np.sin(_T + phi)


class SeasonalDist5d(SeasonalDist):
    def __init__(self, *args, **kwds):
        super(SeasonalDist5d, self).__init__(*args, **kwds)
        self.n_trig_pars = 5
        self.pre_start = [1, 0]

    def trig2pars(self, trig_pars, _T=None):
        if _T is None:
            try:
                _T = self._T
            except AttributeError:
                _T = self._T = (2 * np.pi / 365 * self.doys)[np.newaxis, :]
        try:
            # funny how the obvious thing looks so complicated
            a, b_0, phi, a2, b2 = \
                np.array(trig_pars).reshape(-1, 5).T[:, :, np.newaxis]
            # i lied, this took me some time. but now, the parameters are
            # vertical to T_ and we are broadcasting like some crazy shit
            # pirate radio station
            return a + b_0 * np.sin(_T + phi) * (a2 + b2 * np.sin(_T + phi))
        except ValueError:
            # try with 3 parameters (call happens during get_start_params)
            a, b_0, phi = \
                np.array(trig_pars).reshape(-1, 3).T[:, :, np.newaxis]
            return a + b_0 * np.sin(_T + phi)


class SeasonalDistPar(SeasonalDist):
    def __init__(self, *args, **kwds):
        super(SeasonalDistPar, self).__init__(*args, **kwds)
        self.n_trig_pars = 4
        self.pre_start = [1]

    def trig2pars(self, trig_pars, _T=None):
        """This is the standard "give me distribution parameters for
        trigonometric parameters" that assumes 3 trig pars per distribution
        parameter."""
        if _T is None:
            try:
                _T = self._T
            except AttributeError:
                _T = self._T = (2 * np.pi / 365 * self.doys)[np.newaxis, :]
        try:
            # funny how the obvious thing looks so complicated
            a, b_0, phi, c = \
                np.array(trig_pars).reshape(-1, 4).T[:, :, np.newaxis]
            # i lied, this took me some time. but now, the parameters are
            # vertical to T_ and we are broadcasting like some crazy shit
            # pirate radio station
            return a + b_0 * (1 + np.sin(_T + phi)) ** c
        except ValueError:
            # try with 3 parameters (call happens during get_start_params)
            a, b_0, phi = \
                np.array(trig_pars).reshape(-1, 3).T[:, :, np.newaxis]
            return a + b_0 * np.sin(_T + phi)


class SlidingDist(SeasonalDist):
    """Can we get a better picture at the seasonalities of the distribution
    parameters if we estimate them over a sliding window over the doys?"""
    def __init__(self, distribution, x, dtimes, doy_width=15, order=4,
                 solution=None, *args, **kwds):
        super(SlidingDist, self).__init__(distribution, x, dtimes, *args,
                                          **kwds)
        self.doy_width, self.order = doy_width, order
        # usefull to assess goodness-of-fit/overfitting
        self.n_trig_pars = 2 * order
        self.doys_unique = np.sort(np.unique(self.doys))
        self.n_doys = len(self.doys_unique)

        self.solution = solution
        self._doy_mask = self._sliding_pars = None

    @property
    def doy_mask(self):
        """Returns a (n_unique_doys, len(data)) ndarray"""
        if self._doy_mask is None:
            doy_width, doys = self.doy_width, self.doys
            self._doy_mask = \
                np.empty((len(self.doys_unique), len(self.x)), dtype=bool)
            for doy_i, doy in enumerate(self.doys_unique):
                ii = (doys > doy - doy_width) & (doys <= doy + doy_width)
                if (doy - doy_width) < 0:
                    ii |= doys > (365. - doy_width + doy)
                if (doy + doy_width) > 365:
                    ii |= doys < (doy + doy_width - 365.)
                self._doy_mask[doy_i] = ii
        return self._doy_mask

    @property
    def sliding_pars(self):
        if self._sliding_pars is None:
            if self.verbose:
                progress_bar = my.ProgressBar(self.n_doys)
            self._sliding_pars = \
                np.empty((self.n_doys, (len(self.dist.parameter_names) -
                                        len(self.fixed_pars.keys()))))
            for doy_ii in xrange(self.n_doys):
                data = self.x[self.doy_mask[doy_ii]]
                fixed = {par_name: func(self.doys_unique[doy_ii])
                         for par_name, func in self.fixed_pars.items()}
                self._sliding_pars[doy_ii] = self.dist.fit(data, **fixed)
                if self.verbose:
                    progress_bar.animate(int(doy_ii))
            if self.verbose:
                print
        return self._sliding_pars.T

    @property
    def solution(self):
        if self._solution is None:
            trans = []
            for sl_par in self.sliding_pars:
                trans += [np.fft.rfft(sl_par)]
            self._solution = np.array(trans)
        return self._solution

    @solution.setter
    def solution(self, trans):
        self._solution = trans

    def fourier_approx_new(self, order=4, trig_pars=None):
        if trig_pars is None:
            trig_pars = self.solution

        _fourier_approx = \
            np.empty((len(self.dist.parameter_names), self.n_doys))
        for ii, trans_par in enumerate(trig_pars):
            # find the order biggest amplitudes
            ii_below = \
                np.argsort(np.abs(trans_par))[:len(trans_par) - order - 1]
            pars = np.copy(trans_par)
            pars[ii_below] = 0
            _fourier_approx[ii] = np.fft.irfft(pars, self.n_doys)

        return _fourier_approx

    def fourier_approx(self, order=4, trig_pars=None):
        if trig_pars is None:
            trig_pars = self.solution

        _fourier_approx = \
            np.empty((len(self.dist.parameter_names), self.n_doys))
        for ii, trans_par in enumerate(trig_pars):
            _fourier_approx[ii] = \
                np.fft.irfft(trans_par[:order + 1], self.n_doys)
        return _fourier_approx

    def fit(self, x=None, **kwds):
        if x is not None:
            self.x = x
        return self.solution

    def trig2pars(self, trig_pars, _T=None, order=None):
        order = self.order if order is None else order
        if _T is None:
            try:
                _T = self._T
            except AttributeError:
                _T = self._T = (2 * np.pi / 365 * self.doys)[np.newaxis, :]
        doys = np.atleast_1d(365 * np.squeeze(_T) / (2 * np.pi))
        doys_ii = np.where(my.isclose(self.doys_unique, doys[:, None]))[1]
        return np.array([self.fourier_approx(order, trig_pars)[:, doy_i]
                         for doy_i in doys_ii]).T

    def plot_fourier_fit(self, order=None):
        """Plots the Fourier approximation of all parameters."""
        order = self.order if order is None else order
        fig, axes = plt.subplots(len(self.non_fixed_names), sharex=True,
                                 squeeze=True)
        for par_i, par_name in enumerate(self.non_fixed_names):
            axes[par_i].plot(self.doys_unique, self.sliding_pars[par_i])
            axes[par_i].plot(self.doys_unique,
                             self.fourier_approx_new(order)[par_i],
                             label="new")
            axes[par_i].plot(self.doys_unique,
                             self.fourier_approx(order)[par_i],
                             label="old")
            axes[par_i].set_title("%s Fourier order: %d" % (par_name, order))
        plt.legend(loc="best")
        return fig

if __name__ == "__main__":
    import vg
    dtimes_, met = vg.read_met()
    theta, dtimes = vg.my.sumup(met["theta"], 24, dtimes_)
    theta_dist = SlidingDist(vg.distributions.norm, theta, dtimes)
    theta_dist.fit()
    theta_dist.scatter_pdf()
#    for order in range(5):
#        theta_dist.plot_fourier_fit(order)
#    print theta_dist.chi2_test()
    plt.show()
#    rh, dtimes = vg.my.sumup(met["Qsw"], 24, dtimes_)
#    rh_dist = SlidingDist(vg.distributions.beta, rh, dtimes, 15, 10,
#                          fixed_pars=vg.conf.par_known["Qsw"])
#    rh_dist.fit()
    #rh_dist.plot_fourier_fit(order=10)
    #rh_dist.scatter_pdf()


#    vg.plt.plot(qsw_dist.sliding_pars)
#    qsw_dist.plot_seasonality_fit()
#    vg.plt.show()
