import warnings
import numpy as np
from scipy import optimize, interpolate
import matplotlib.pylab as plt
import distributions
from vg import times
from vg.smoothing import smooth
from vg import helpers as my


def apply_kernel_ne(kernel_width, data, eval_points=None):
    if eval_points is None:
        distances = distance_array(data, data)
    else:
        distances = distance_array(data, eval_points)
    pi = np.pi
    densities = ne.evaluate(
        "1.0 / (sqrt(2 * pi) * kernel_width) * " +
        "exp(-distances ** 2 / (2 * kernel_width ** 2))")
    return densities


def apply_kernel_np(kernel_width, data, eval_points=None):
    if eval_points is None:
        distances = distance_array(data, data)
    else:
        distances = distance_array(data, eval_points)
    return (1.0 / (np.sqrt(2 * np.pi) * kernel_width) *
            np.exp(-distances ** 2 / (2 * kernel_width ** 2)))

try:
    import numexpr as ne
    apply_kernel = apply_kernel_ne
except ImportError:
    warnings.warn("Could not import numexpr. Using numpy for KDE instead.")
    apply_kernel = apply_kernel_np


def array_gen(scalar):
    return lambda tt: scalar * np.ones_like(np.atleast_1d(tt))


def distance_array(x_vec, y_vec):
    """creating a matrix with residuals (x:horizontal,y:vertical)"""
    x_vec, y_vec = map(np.atleast_1d, (x_vec, y_vec))
    return x_vec[np.newaxis, :] - y_vec[:, np.newaxis]


def doy_distance(doy1, doy2):
    return np.min([np.abs(doy2 - doy1),
                   np.abs(doy1 - doy2 - 365),
                   np.abs(doy2 - doy1 - 365)], axis=0)


class SeasonalKDE(distributions.MyDist):
    def __init__(self, data, datetimes, solution=None, doy_width=15, nx=1e3,
                 fixed_pars=None, verbose=True):
        """
        nx is the number of data points used to estimate the quantiles.
        lower- and upper bound must be functions that accept a doy. If None
        seasonal minima and maxima of data will be used.

        Parameters
        ----------
        data : 1dim ndarray
            may contain nans (but no nuts!)
        datetimes : sequence of datetime objects
        solution : sequence
            contains kernel widths and quantile_grid as returned from
            SeasonalKDE.fit
        doy_width :  int
            kernel width in time direction.
        nx : int
            number of datapoints for each doy to base the pdf, cdf and ppf
            estimates on
        fixed_pars : dict
            with 'l' and or 'u' mapping to a function that accepts doys and
            returns lower and upper bounds.
        verbose : boolean
        """
        finite_ii = np.isfinite(data)
        self.data = data[finite_ii]
        self.doy_width = doy_width
        self.nx = nx
        self.verbose = verbose
        self.doys = times.datetime2doy(datetimes[finite_ii])

        # those will be populated by cached properties
        self._doy_mask = self._doy_mask_dict = self._kernel_widths = None
        self._density_grid = self._x_grid = self._quantile_grid = None
        self._cdf_interp_per_day = self._ppf_interp_per_day = None
        self._data_bounds = self._lower_bound = self._upper_bound = None

        if solution is not None:
            self.solution = solution

        self.doys_unique = np.sort(np.unique(self.doys))
        if fixed_pars is None:
            fixed_pars = {}

        try:
            fixed_pars["l"] = fixed_pars["lc"]
        except KeyError:
            pass
        if "l" in fixed_pars:
            self.lower_bound = fixed_pars["l"]
        try:
            fixed_pars["u"] = fixed_pars["uc"]
        except KeyError:
            pass
        if "u" in fixed_pars:
            self.upper_bound = fixed_pars["u"]

        upper = self.upper_bound(self.doys)
        upper_ii = self.data >= upper
        if np.any(upper_ii):
            self.data[upper_ii] = (np.sign(upper[upper_ii]) *
                                   np.abs(upper[upper_ii]) * (1 - 1e-3))

        lower = self.lower_bound(self.doys)
        lower_ii = self.data <= lower
        if np.any(lower_ii):
            self.data[lower_ii] = (np.sign(lower[lower_ii]) *
                                   np.abs(lower[lower_ii]) * (1 + 1e-3))

    @property
    def data_bounds(self):
        """Here for optimization: We do not let lower and upper bound make
        a long loop over the data for min and max resp. We do it here once and
        cache it."""
        if self._data_bounds is None:
            exact_doy_mask = self.doys == self.doys_unique[:, np.newaxis]
            self._data_bounds = \
                np.array([(self.data[doy_ii].min(), self.data[doy_ii].max())
                          for doy_ii in exact_doy_mask])
        return self._data_bounds

    def lower_bound(self, doy):
        if self._lower_bound is None:
            data_mins, data_maxs = self.data_bounds.T
            data_mins_smooth = \
                smooth(data_mins, 5 * self.doy_width, periodic=True)
            # we have to lower the smoothed mins below the actual data mins
            diffs = data_mins_smooth - data_mins
            # ideal would be a kernel width, but we do not have those yet
            diff_thresh = ((data_maxs.mean() - data_mins.mean()) /
                           (len(self.data) / 365))
            while np.any(diffs > -diff_thresh):
                sub = np.zeros_like(diffs)
                argmax = np.argmax(diffs)
                diff_max = diffs[argmax]
                sub[argmax] = diff_max
                sub = smooth(sub, 5 * self.doy_width, periodic=True)
                sub *= (1 + diff_thresh) * 5 * self.doy_width / diff_max
                data_mins_smooth -= sub
                diffs = data_mins_smooth - data_mins
            self._lower_bound = data_mins_smooth
        doy = np.asarray(doy, int) - 1
        return self._lower_bound[doy]

    def upper_bound(self, doy):
        if self._upper_bound is None:
            data_mins, data_maxs = self.data_bounds.T
            data_maxs_smooth = \
                smooth(data_maxs, 5 * self.doy_width, periodic=True)
            # we have to lift the smoothed maxs above the actual data maxs
            diffs = data_maxs - data_maxs_smooth
            # ideal would be a kernel width, but we do not have those yet
            diff_thresh = ((data_maxs.mean() - data_mins.mean()) /
                           (len(self.data) / 365))
            while np.any(diffs > -diff_thresh):
                add = np.zeros_like(diffs)
                argmax = np.argmax(diffs)
                diff_max = diffs[argmax]
                add[argmax] = diff_max
                add = smooth(add, 5 * self.doy_width, periodic=True)
                add *= (1 + diff_thresh) * 5 * self.doy_width / diff_max
                data_maxs_smooth += add
                diffs = data_maxs - data_maxs_smooth
            self._upper_bound = data_maxs_smooth
        doy = np.asarray(doy, int) - 1
        return self._upper_bound[doy]

    @property
    def doy_mask(self):
        """Returns a (n_unique_doys, len(data)) ndarray"""
        if self._doy_mask is None:
            doy_width, doys = self.doy_width, self.doys
            self._doy_mask = \
                np.empty((len(self.doys_unique), len(self.data)), dtype=bool)
            for doy_i, doy in enumerate(self.doys_unique):
                ii = (doys > doy - doy_width) & (doys <= doy + doy_width)
                if (doy - doy_width) < 0:
                    ii |= doys > (365. - doy_width + doy)
                if (doy + doy_width) > 365:
                    ii |= doys < (doy + doy_width - 365.)
                self._doy_mask[doy_i] = ii
        return self._doy_mask

    @property
    def doy_mask_dict(self):
        if self._doy_mask_dict is None:
            self._doy_mask_dict = {}
            for row_i, row in enumerate(self.doy_mask):
                self._doy_mask_dict[self.doys_unique[row_i]] = row
        return self._doy_mask_dict

    @property
    def x_grid(self):
        if self._x_grid is None:
            xx = np.empty((len(self.doys_unique), self.nx))
            for doy_i, doy in enumerate(self.doys_unique):
                xx[doy_i] = np.linspace(self.lower_bound(doy),
                                        self.upper_bound(doy), self.nx)
            self._x_grid = xx
        return self._x_grid

    @property
    def density_grid(self):
        if self._density_grid is None:
            densities = np.empty_like(self.x_grid)
            for doy_i, doy in enumerate(self.doys_unique):
                kernel_width = self.kernel_widths[doy_i]
                ii = self.doy_mask_dict[doy]
                x = self.x_grid[doy_i]
                densities[doy_i] = \
                    self.density_per_doy(kernel_width, self.data[ii],
                                         self.doys[ii], doy, x)
            self._density_grid = densities
            #HACK the following will normalize the densities
            self.quantile_grid
        return self._density_grid

    @property
    def quantile_grid(self):
        if self._quantile_grid is None:
            quantiles = my.cumtrapz(y=self.density_grid, x=self.x_grid,
                                    axis=1, initial=0)
            # this normalizes the quantiles
            quantiles *= (quantiles[:, -1] ** -1)[:, np.newaxis]
            # just making it sure...
            quantiles[:, 0] = 0
            quantiles[:, -1] = 1.
            self._quantile_grid = quantiles
        return self._quantile_grid

    @quantile_grid.setter
    def quantile_grid(self, grid):
        self._quantile_grid = grid

    @property
    def cdf_interp_per_day(self):
        if self._cdf_interp_per_day is None:
            self._cdf_interp_per_day = \
                [interpolate.interp1d(self.x_grid[doy_i], quantiles)
                 for doy_i, quantiles in enumerate(self.quantile_grid)]
        return self._cdf_interp_per_day

    @property
    def ppf_interp_per_day(self):
        if self._ppf_interp_per_day is None:
            self._ppf_interp_per_day = \
                [interpolate.interp1d(quantiles, self.x_grid[doy_i])
                 for doy_i, quantiles in enumerate(self.quantile_grid)]
        return self._ppf_interp_per_day

    def density_per_doy(self, kernel_width, data, doys, doy_middle,
                        eval_data=None, leave_one_out=False):
        densities = apply_kernel(kernel_width, data, eval_data)
        doy_scale = (1 - doy_distance(doy_middle, doys) /
                     (self.doy_width + 1)) / (self.doy_width - 1)
        densities *= doy_scale
        # we want the densities at a specific doy to integrate to one
        densities *= len(doys) / self.doy_width
        if leave_one_out:
            # sets main diagonal to 0 (exploiting the fact that ravel returns
            # a view)
            densities.ravel()[::len(data) + 1] = 0
        return np.sum(densities, axis=1) / float(len(densities) - 1)

    def optimize_kernelwidth(self, kernel_width, data, doys, doy_middle):
        """optimizing kernel width with MLM leave one out"""
        densities = self.density_per_doy(np.abs(kernel_width), data, doys,
                                         doy_middle, leave_one_out=True)
        return -np.nansum(np.log(densities))

    def fit(self):
        """This estimates the kernel widths per doy AND returns the
        interpolating functions for cdf and ppf."""
        kernel_widths = np.empty(len(self.doys_unique))
        if self.verbose:
            progress_bar = my.ProgressBar(len(self.doys_unique))
        for doy_i, doy in enumerate(self.doys_unique):
            ii = self.doy_mask[doy_i]
            if doy_i == 0:
                x0 = (self.data[ii].max() - self.data[ii].min()) / 10.
            else:
                x0 = kernel_widths[doy_i - 1]
            kernel_width = \
                optimize.fmin(self.optimize_kernelwidth, x0,
                              args=(self.data[ii], self.doys[ii], doy),
                              disp=False)
            kernel_widths[doy_i] = kernel_width
            if self.verbose:
                progress_bar.animate(int(doy_i))
        if self.verbose:
            print
        kernel_widths = np.abs(kernel_widths)
        kernel_widths = smooth(kernel_widths, self.doy_width, periodic=True)
        self._kernel_widths = kernel_widths
        return kernel_widths, self.quantile_grid

    @property
    def kernel_widths(self):
        if self._kernel_widths is None:
            self.fit()
        return self._kernel_widths

    @kernel_widths.setter
    def kernel_widths(self, widths):
        self._kernel_widths = widths

    @property
    def solution(self):
        return self.kernel_widths, self.quantile_grid

    @solution.setter
    def solution(self, solution):
        self.kernel_widths, self.quantile_grid = solution

    def pdf(self, solution, x, doys):
        x, doys = map(np.atleast_1d, (x, doys))
        densities = np.empty((len(doys), len(x)))
        for doy in doys:
            doy_i = np.where(my.isclose(self.doys_unique, doy))[0][0]
            kernel_width = self.kernel_widths[doy_i]
            ii = self.doy_mask_dict[doy]
            densities[doy_i] = self.density_per_doy(kernel_width,
                                                    self.data[ii],
                                                    self.doys[ii], doy, x)
        return densities

    def cdf(self, solution=None, x=None, doys=None):
        if solution is not None:
            self.solution = solution
        if x is None:
            x = self.data
        if doys is None:
            doys = self.doys

        x, doys = map(np.atleast_1d, (x, doys))
        quantiles = np.nan * np.ones_like(x)
        for ii, (doy, x_single) in enumerate(zip(doys, x)):
            doy_i = np.where(my.isclose(self.doys_unique, doy))[0][0]
            try:
                quantiles[ii] = self.cdf_interp_per_day[doy_i](x_single)
            except ValueError, e:
                if x_single < self.lower_bound(doy):
                    #TODO: this is very macguyver
                    quantiles[ii] = 1e-6
                    continue
                elif x_single > self.upper_bound(doy):
                    quantiles[ii] = 1 - 1e-6
                    continue
                raise e
        return quantiles

    def ppf(self, solution, quantiles, doys):
        if solution is not None:
            self.solution = solution
        quantiles, doys = map(np.atleast_1d, (quantiles, doys))
        x = np.empty_like(quantiles)
        for ii, (doy, quantile_single) in enumerate(zip(doys, quantiles)):
            doy_i = np.where(my.isclose(self.doys_unique, doy))[0][0]
            x[ii] = self.ppf_interp_per_day[doy_i](quantile_single)
        return x

    def scatter_pdf(self, solution, n_sumup=24, figsize=None, title=None,
                     opacity=.25, plot_kernel_width=False):
        doys = \
            self.doys_unique.repeat(self.nx).reshape(len(self.doys_unique), -1)
        fig = plt.figure(figsize=figsize)
        ax1 = fig.gca()
        co = ax1.contourf(doys, self.x_grid / n_sumup, self.density_grid, 15)
#        plt.colorbar(co)
        plt.scatter(self.doys, self.data / n_sumup, marker="o",
                   facecolors=(0, 0, 0, 0), edgecolors=(0, 0, 0, opacity))
#        plt.plot(self.doys_unique, self.lower_bound(self.doys_unique), "r",
#                 label="lower bound")
#        plt.plot(self.doys_unique, self.upper_bound(self.doys_unique), "r",
#                 label="upper bound")
        ax1.set_ylim(self.x_grid.min() / n_sumup, self.x_grid.max() / n_sumup)

        if plot_kernel_width:
            ax2 = plt.gca().twinx()
            ax2.plot(self.doys_unique, self.kernel_widths, "r-",
                     label="Kernel width")
            ax2.set_ylim(self.kernel_widths.min(), self.kernel_widths.max())

        plt.xlim(0, 366)
        plt.xticks((1, 32, 60, 91, 121, 152, 182, 213, 244, 274, 305, 335),
                   ('Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug',
                    'Sep', 'Oct', 'Nov', 'Dec'), rotation=45)

#        plt.xlabel("Day of Year")
        plt.grid()
        #plt.legend()
        if title is not None:
            plt.title(title)
        #plt.tight_layout()
        return fig

    def scatter_cdf(self, solution, n_sumup=1, figsize=None, title=None):
        doys = self.doys_unique.repeat(1e3).reshape(len(self.doys_unique), 1e3)
        fig = plt.figure(figsize=figsize)
        ax1 = fig.gca()
        co = ax1.contourf(doys, self.x_grid, self.quantile_grid, 15)
        plt.colorbar(co)
        plt.scatter(self.doys, self.data / n_sumup, marker="o",
                   facecolors=(0, 0, 0, 0), edgecolors=(0, 0, 0, .5))
#        plt.plot(self.doys_unique, self.lower_bound(self.doys_unique), "r",
#                 label="lower bound")
#        plt.plot(self.doys_unique, self.upper_bound(self.doys_unique), "r",
#                 label="upper bound")
        ax1.set_ylim(self.x_grid.min(), self.x_grid.max())
        ax2 = plt.gca().twinx()
        ax2.plot(self.doys_unique, self.kernel_widths, label="Kernel width")
        ax2.set_ylim(self.kernel_widths.min(), self.kernel_widths.max())

        plt.xlim(0, len(doys))
        plt.xlabel("Day of Year")
        plt.grid()
        plt.legend()
        if title is not None:
            plt.title(title)
        return fig


if __name__ == "__main__":
    import vg
    import shelve
    dtimes_, met = vg.read_met()
    u, dtimes = vg.my.sumup(met["u"], 24, dtimes_)
    sh = shelve.open(vg.conf.seasonal_cache_file, "r")
    seas_class, dist_class, solution = sh["u"]
    dist = seas_class(u, dtimes, solution)
#    dist = SeasonalKDE(u, dtimes, doy_width=25)
#    dist.fit()
    dist.scatter_pdf(None)
    plt.show()
    quantiles = dist.cdf(solution, x=u, doys=vg.times.datetime2doy(dtimes))
