import numpy as np
import scipy as sp
import matplotlib.pyplot as plt
from vg import helpers as my


def time_part_average(values, times, time_format_str="%w", expand=True, 
                      return_times=False):
    """Return averages per time part, e.g. monthly means."""
    time_parts = times.time_part(times, time_format_str)
    time_parts_diffs = time_parts[1:] - time_parts[:-1]
    breaks = np.where(time_parts_diffs != 0)[0] + 1
    values_splitted = np.split(values, breaks)
    values_means = np.array([val.mean() for val in values_splitted])
    if expand:
        repeats = [len(val) for val in values_splitted]
        time_part_average.repeats = repeats
        values_means = np.repeat(values_means, repeats)
    if return_times:
        return values_means, np.r_[[times[0]], times[breaks]]
    return values_means


def matr_img(array, title=None, xlabels=None, ylabels=None, fig=None, 
             figsize=None, fontsize=12, colorbar=False, *args, **kwds):
    fig = plt.figure(figsize=figsize) if fig is None else fig
    plt.imshow(array, interpolation="nearest", **kwds)
#    fig.canvas.mpl_connect("draw_event", my.scale_yticks)
    if xlabels is not None:
        plt.xticks(np.arange(len(xlabels)), xlabels, rotation=45)
    if ylabels is not None:
        plt.yticks(np.arange(len(ylabels)), ylabels)
    if title is not None:
        plt.title(title)
        fig.canvas.set_window_title("%s (%d)" %
                                    (title, fig.canvas.manager.num))
    ax = plt.gca()
    ax.tick_params(left="off", bottom="off", top="off", right="off")
    
    if colorbar:
        plt.colorbar()
    for x in range(array.shape[0]):
        for y in range(array.shape[1]):
            # do not let yourself be fooled by the strange imshow-coordinate
            # system
            plt.text(y, x, "%.2f" % array[x, y], fontsize=fontsize,
                     horizontalalignment='center', verticalalignment="center")
    return fig


def corr_img(data, k=0, title=None, var_names=None, rank=False, *args, 
             **kwds):
    """Plots the correlation matrix of data using imshow"""
    if rank:
        corr = cross_rank_corr(data, k)
    else:
        corr = cross_corr(data, k)
    return matr_img(corr, title, xlabels=var_names, ylabels=var_names,
                    vmin= -1, vmax=1, *args, **kwds)


def auto_cov(data, k):
    """Return the autocovariance-vector for lag k. Variables are assumed to
    be stored in rows, with time extending across the columns."""
    if np.rank(data) == 1:
        data = data[np.newaxis, :]
    n_vars, n_data = data.shape
    k = np.atleast_1d(k)
    means = np.mean(data, axis=1).reshape((n_vars, 1))
    auto = np.empty((len(k), n_vars))
    for ii, lag in enumerate(k):
        if lag == 0:
            auto[ii] = np.mean((data - means) ** 2, axis=1)
        else:
            auto[ii] = np.mean((data[:, :-lag] - means) *
                               (data[:, lag:] - means), axis=1)
            auto[ii] *= float(n_data - lag) / n_data
    return np.squeeze(auto)


def auto_corr(data, k):
    """Return the autocorrelation-vector for lag k. Variables are assumed to
    be stored in rows, with time extending across the columns."""
    return np.squeeze(auto_cov(data, k) / auto_cov(data, 0))


def _partial_autocorr_uni(data, k):
    """Univariate implementation. Call partial_autocorr for multi- and
    univariate cases instead.
    Autocorrelation with the influence of the time steps between t and t-k
    removed. See Hartung p.677f. Tested according to example on p.694"""
    k = np.atleast_1d(k)
    autos = np.empty(len(k))
    r_k = auto_corr(data, 1)
    for lag_i, lag in enumerate(k):
        if lag == 0:
            autos[lag_i] = 1
        elif lag == 1:
            autos[lag_i] = r_k
        else:
            # the following block was written for k as an integer. optimization
            # in terms of not executing stuff twice is probably possible
            # read P_k-1
            P_km1 = np.atleast_1d(auto_corr(data, range(1, lag)))
            P_km1_rev = P_km1[::-1]
            # read P_k-1,k-1
            P_km1km1 = np.matrix(np.ones((lag - 1, lag - 1)))
            for ii in range(lag - 1):
                if ii > 0:
                    P_km1km1[ii, :ii] = P_km1_rev[-ii:]
                if ii < lag - 2:
                    P_km1km1[ii, ii + 1:] = P_km1[ii:-1]
            P_km1km1_inv = P_km1km1.I
            P_km1, P_km1_rev = np.matrix(P_km1), np.matrix(P_km1_rev)
            autos[lag_i] = ((r_k - P_km1 * P_km1km1_inv * P_km1_rev.T) /
                            np.sqrt((1 - P_km1 * P_km1km1_inv * P_km1.T) *
                                    (1 - P_km1_rev * P_km1km1_inv *
                                     P_km1_rev.T)))
    return np.squeeze(autos)


def partial_autocorr(data, k):
    """Autocorrelation with the influence of the time steps between t and t-k
    removed. See Hartung p.677f. Tested according to example on p.694"""
    if np.rank(data) == 1:
        return _partial_autocorr_uni(data, k)
    else:
        return np.array([_partial_autocorr_uni(var, k) for var in data])


def plot_auto_corr(data, k_range=7 * 24, title="", var_names=None,
                   n_per_day=1, partial=False, n_difference=0,
                   draggable=False, figsize=None, window_title=True, *args,
                   **kwds):
    fig = plt.figure(figsize=figsize)
    fig.canvas.mpl_connect('draw_event', my.scale_yticks)
    lags = np.arange(k_range)
    corr_func = partial_autocorr if partial else auto_corr
    d_data = np.copy(data)
    if np.rank(data) == 1:
        d_data = d_data[np.newaxis, :]
    if np.rank(data) == 2:
        d_data = d_data[np.newaxis, ...]
    for data_2d, linestyle in zip(d_data, ("-", "--", ":")):
        for diff in range(n_difference + 1):
            fig.add_subplot(n_difference + 1, 1, diff + 1)
    #        autos = np.array([corr_func(d_data, k) for k in lags])
            autos = corr_func(data_2d, lags)
            d_data = data_2d[:, 1:] - data_2d[:, :-1]
            for auto, color in zip(autos.T, "bgrcmykw"):
                plt.plot(auto, color + linestyle + "x")
            xes = range(0, k_range, n_per_day)
            plt.xticks(xes, xes)
            plt.grid()
            if var_names is not None:
                leg = plt.legend(var_names, loc="center left", 
                                 bbox_to_anchor=(1, .5), frameon=False,
                                 borderaxespad=0.)
                leg.draggable(draggable)
    plt.suptitle(title)
    if window_title:
        fig.canvas.set_window_title("%s (%d)" %
                                    (title, fig.canvas.manager.num))
    return fig


#def cross_cov_nan(data, k):
#    """Biased cross covariance of an array containig nans."""
#    n_vars = data.shape[0]
#    means = my.nanavg(data, axis=1)[:, np.newaxis]
#    cross = np.empty((n_vars, n_vars))
#    for ii in range(n_vars):
#        for jj in range(n_vars):
#            cross[ii, jj] = my.nanavg((data[ii, :-k
def cross_cov(data, k, means=None):
    """Return the cross-covariance matrix for lag k. Variables are assumed to
    be stored in rows, with time extending across the columns."""
#    if k == 0:
#        # very funny, jackass
#        return np.cov(data)
    n_vars = data.shape[0]
    k_right = -abs(k) if k else None
    if means is None:
        means = my.nanavg(data, axis=1).reshape((n_vars, 1))
        ddof = 1
    else:
        ddof = 0
    cross = np.empty((n_vars, n_vars))
    for ii in range(n_vars):
        for jj in range(n_vars):
            cross[ii, jj] = \
                my.nanavg((data[ii, :k_right] - means[ii]) *
                          (data[jj, k:] - means[jj]), ddof=ddof)
    return cross


def cross_corr(data, k):
    """Return the cross-correlation-coefficient matrix for lag k. Variables are
    assumed to be stored in rows, with time extending across the columns."""
    stds = np.std(data, axis=1)[:, np.newaxis]
    stds_dot = stds * stds.T  # dyadic product of row-vector stds
    return cross_cov(data, k) / stds_dot


def partial_corr_(x, y, u, rank_cor=True):
    """Is the correlation between x and y because of u?.
    Note that this is only a reasonable measure if x, y and u are normally
    distributed."""
    if rank_cor:
        corr_func = my.rank_corrcoef
    else:
        corr_func = lambda x, y: np.corrcoef(x, y)[0, 1]
    r_xy = corr_func(x, y)
    r_xu = corr_func(x, u)
    r_yu = corr_func(y, u)
    return (r_xy - r_xu * r_yu) / ((1 - r_xu ** 2) * (1 - r_yu ** 2)) ** .5


def rank_corr_ij(data):
    """Computes the rank correlation coefficient matrix."""
    n_vars = data.shape[0]
    r_ij = np.empty((n_vars, n_vars))
    # fill the lower triangular matrix
    for ii in range(n_vars):
        for jj in range(ii):
            r_ij[ii, jj] = my.rank_corrcoef(data[ii], data[jj])
        # do not compute the correlation of a variable with itself
        r_ij[ii, ii] = 1
    # fill the upper triangular matrix, exploiting symmetry
    for ii in range(n_vars):
        if ii + 1 < n_vars:
            for jj in range(ii + 1, n_vars):
                r_ij[ii, jj] = r_ij[jj, ii]
    return r_ij


def cross_rank_corr(data, k):
    """Computes the cross rank correlation for lag k."""
    if k == 0:
        # this should be much faster because it avoids doing double work
        return rank_corr_ij(data)
    n_vars = data.shape[0]
    r_ij = np.empty((n_vars, n_vars))
    for ii in range(n_vars):
        for jj in range(n_vars):
            r_ij[ii, jj] = my.rank_corrcoef(data[ii, k:], data[jj, :-k])
    return r_ij


def partial_corr(data, index, diff=False, rank_cor=True, r_ij=None):
    """Returns a matrix with partial (rank) correlation coefficients of all
    variables with respect to the variables in row "index".
    If diff is given as True, the difference to the correlation coefficients
    will be returned.
    If rank_cor is given as False, the ordinary partial correlations are used.
    """
    if r_ij is None:
        if rank_cor:
            r_ij = rank_corr_ij(data)
        else:
            r_ij = cross_corr(data, 0)
    # correlation coefficients between all variables and the one under study
    r_ik = r_ij[index]
    # the same as a column array
    r_kj = r_ik[:, np.newaxis]
    # to iterate is human and to recurse is divine, but to broadcast is to
    # enter Nirwana
    partial = (r_ij - r_ik * r_kj) / ((1 - r_ik ** 2) * (1 - r_kj ** 2)) ** .5
    if diff:
        partial = r_ij - partial
    return partial


def hurst_coefficient(values):
    """Estimates the Hurst coefficient.

    Parameters
    ----------
    values :    1-dim ndarray
                Data representing a time-series.

    References
    ----------
    http://en.wikipedia.org/wiki/Hurst_coefficient#Estimating_the_exponent
    """
    # find n_partitions, so that the minimum number of observations in a
    # sub-time-series is 8
    N = values.size
    n_partitions = int(np.log2(float(N) / 8))
    Rn_Sn = np.zeros(n_partitions)
    region_sizes = np.zeros(n_partitions)
    for part_i in range(n_partitions):
        # form sub-time-series by reshaping, so we can do all calculation with
        # the help of axis=1
        n_parts = 2 ** part_i
        cutoff = -(N % n_parts)
        if cutoff == 0:
            cutoff = None
        Xt = values[:cutoff].reshape((n_parts, -1))
        Yt = Xt - Xt.mean()
        Zt = np.cumsum(Yt, axis=1)
        Rn = np.max(Zt, axis=1) - np.min(Zt, axis=1)
        #Sn = np.sqrt(np.mean((Xt - Xt.mean(axis=1)[:, np.newaxis]) ** 2))
        Sn = np.std(Xt, axis=1)
        Rn_Sn[part_i] = np.mean(Rn / Sn)
        region_sizes[part_i] = Xt.shape[1]

    # do a regression on x=log(n), y=log(Rn/Sn)
    return sp.stats.linregress(np.log2(region_sizes), np.log2(Rn_Sn))[0]


#def test_partial_corr(data, index, alpha=.05, rank_cor=True, r_ij=None):
#    """Returns an array with True/False values whether the Null Hypothesis
#    that x and y are partially uncorrelated can be rejected.
#    See Hartung p.562"""
#    n = data.shape[1]
#    if r_ij is None:
#        r_ij = rank_corr_ij(data)
#    # to only test for partially uncorrelatedness where it makes sense, we
#    # limit ourselves to elements that are rank-correlated at all
#    # see Hartung p.560
#    corr_test_statistic = n * (n - 1) / 2. * r_ij
#    # i have no theoretical values of the test statistic, but the following
#    # transformation puts it into the standard normal world.
#    corr_test_statistic /= (n * (n - 1) * (2. * n + 5) / 18) ** .5
#    corr_mask = (np.abs(corr_test_statistic) >
#                 sp.stats.norm.ppf(1 - .5 * alpha))
#    # approximation of the theoretical K-statistic
#    partial = partial_corr(data, index, rank_cor=rank_cor, r_ij=r_ij)
#    test_statistic = np.abs(partial * (n - 3) ** .5 /
#                            (1 - partial ** 2) ** .5)
##    return corr_mask & (test_statistic > stats.t.ppf(1 - .5 * alpha, n - 3))
#    return test_statistic > sp.stats.t.ppf(1 - .5 * alpha, n - 3)
